'''
EEGNet-8,2

References
----------
V. J. Lawhern, A. J. Solon, N. R. Waytowich, S. M. Gordon, C. P. Hung, and B. J. Lance,
“EEGNet: a compact convolutional neural network for EEG-based brain–computer interfaces,”
J. Neural Eng., vol. 15, no. 5, p. 056013, Jul. 2018, doi: 10.1088/1741-2552/aace8c.
'''


import torch
import torch.nn as nn

from .utils import Conv2dWithNorm, LinearWithNorm


class EEGNet(nn.Module):
    def __init__(self, nCh=22, nTime=1000, C1=63, F1=8, D=2, F2=16, C2=15, 
                 P1=8, P2=16, p=0.5, cls=4) -> None:
        super().__init__()

        self.filter = nn.Sequential(
            nn.Conv2d(1, F1, (1, C1), padding=(0, C1//2), bias=False),
            nn.BatchNorm2d(F1)
        )

        self.depthwise_conv = nn.Sequential(
            Conv2dWithNorm(F1, D*F1, (nCh, 1), groups=F1, bias=False, max_norm=1),
            nn.BatchNorm2d(D*F1),
            nn.ELU(),
            nn.AvgPool2d((1, P1)),
            nn.Dropout(p)
        )

        self.separable_conv = nn.Sequential(
            nn.Conv2d(D*F1, D*F1, (1, C2), padding=(0, C2//2), groups=D*F1, bias=False),
            nn.Conv2d(D*F1, F2, 1, bias=False),
            nn.BatchNorm2d(F2),
            nn.ELU(),
            nn.AvgPool2d((1, P2)),
            nn.Dropout(p)
        )

        self.flatten = nn.Flatten()
        self.fc = nn.Sequential(
            # Experimental results show that using linearwithnorm will lead to 
            # performance degradation.
            # LinearWithNorm(self.get_size(nCh, nTime), cls, bias=True, max_norm=0.25)
            nn.Linear(self.get_size(nCh, nTime), cls, bias=True),
            nn.LogSoftmax(dim=1)
        )

    def get_size(self, nCh, nTime):
        x = torch.randn(1, 1, nCh, nTime)
        out = self.filter(x)
        out = self.depthwise_conv(out)
        out = self.separable_conv(out)
        return self.flatten(out).size(1)

    def forward(self, x):
        out = self.filter(x)
        out = self.depthwise_conv(out)
        out = self.separable_conv(out)
        out = self.flatten(out)
        return self.fc(out)


if __name__ == '__main__':
    from torchinfo import summary
    net = EEGNet(22, 1000).cuda()
    summary(net, (1, 1, 22, 1000))


#!/usr/bin/env python
# coding: utf-8

"""
    Util functions for dpeeg.
    
    @Author  : SheepTAO
    @Time    : 2023-09-15
"""


import os
import mne
import torch
import random
import inspect
import functools
import numpy as np
from .tools.logger import Logger
from typing import (
    Any,
    List,
    Union,
    Callable,
    Iterable,
    TypeVar,
)


DPEEG_SEED : int = 42               # Life, the Universe, and Everything
DPEEG_LOGER_LEVEL = 'INFO'
DPEEG_DIR = os.path.join(os.path.expanduser('~'), 'dpeeg')


loger = Logger('dpeeg', flevel=None)


# Provide help for static type checkers:
# https://mypy.readthedocs.io/en/stable/generics.html#declaring-decorators
_FuncT = TypeVar('_FuncT', bound=Callable[..., Any])


def verbose(func : _FuncT) -> _FuncT:
    '''Verbose decorator to allow functions to override log-level.

    Parameters
    ----------
    func : callable
        Function to be decorated by setting the verbosity level.

    Returns
    -------
    dec : callable
        The decorated function.
    '''
    @functools.wraps(func)
    def inner(*args, **kwargs) -> _FuncT:
        level = kwargs.get('verbose', DPEEG_LOGER_LEVEL)
        if level:
            loger.update_sh_level(level)
        else:
            loger.update_sh_level(DPEEG_LOGER_LEVEL)
        # for debug
        # print(level, func)
        return func(*args, **kwargs)
    return inner # type: ignore


def set_log_level(
    verbose : Union[int, str] = DPEEG_LOGER_LEVEL,
    ret_old_level : bool = False
) -> Union[None, str]:
    '''Set the global logging level.
    
    Parameters
    ----------
    verbose : int, str, optional
        The verbosity of messages to print. If a str, it can be either DEBUG,
        INFO, WARNING, ERROR, or CRITICAL. Note that these are for convenience
        and are equivalent to passing in logging.DEBUG, etc. If None, the 
        environment variable `DPEEG_LOGER_LEVEL` is read, defaults to INFO.
    ret_old_level : bool, optional
        If True, return the old verbosity level.

    Returns
    -------
    old_level : str, None
        The old level. Only returned if `ret_old_level` is True.
    '''
    global DPEEG_LOGER_LEVEL
    old_level = loger.get_sh_level()
    DPEEG_LOGER_LEVEL = verbose
    mne.set_log_level(verbose)
    if ret_old_level:
        return old_level


def _set_random_seed(seed : int) -> None:
    '''Set the seed for Python's built-in random module and numpy.

    Parameters
    ----------
    seed: int
        The random seed to use.
    '''
    random.seed(seed)
    np.random.seed(seed)


def _set_torch_seed(seed : int) -> None:
    '''Set the seed for Python's built-in random module and numpy.

    Parameters
    ----------
    seed: int
        The random seed to use.

    '''
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)


def set_seed(
    seed : int = DPEEG_SEED, 
    ret_old_seed : bool = False
) -> Union[None, int]:
    '''Set the seed for random, numpy and PyTorch.

    Parameters
    ----------
    seed : int
        The random seed to use. If None, the environment variable `DPEEG_SEED` 
        is read, defaults to 42.
    ret_old_seed : bool, optional
        If True, return the old seed.

    Returns
    -------
    old_seed : int, None
        The old seed. Only returned if `ret_old_seed` is True.
    '''
    global DPEEG_SEED
    old_seed = DPEEG_SEED
    DPEEG_SEED = seed
    _set_random_seed(seed)
    _set_torch_seed(seed)
    if ret_old_seed:
        return old_seed


def unpacked(*args) -> list:
    '''Positional arguments are unpacked as lists.
    '''
    result = [X for X in args]
    return result


def dict_to_str(kwargs : dict, symbol : str = ', ') -> str:
    '''Convert the dictionary into a string format.

    Parameters
    ----------
    kwargs : dict
        The dictionary to be converted.
    symbol : str
        Join all key-value pairs with the specified separator character.
    '''
    s = [f'{k}={v}' for k, v in kwargs.items()]
    return symbol.join(s)


def align_text(prefix : str, text : str, only_head : bool = True) -> str:
    '''Aligns text, adding a specified prefix to the beginning of line.

    Parameters
    ----------
    prefix : str
        The prefix to be added at the beginning of line.
    text : str
        The text to be aligned, including multiple lines.
    only_head : bool
        If false, prefix will be prepended to each line. Otherwise, spaces with
        a length of prefix will be added to other lines except the first line.

    Returns
    -------
    str: The aligned text.
    '''
    lines = text.splitlines()
    aligned_lines = [f'{prefix}{line}' if i == 0 and only_head else 
                f'{" " * len(prefix)}{line}' for i, line in enumerate(lines)]
    return '\n'.join(aligned_lines)


def get_init_args(
    obj,
    locals : dict[str, Any],
    format : str = 'log',
) -> str:
    '''Get object initialization parameters.

    Assemble the parameters passed in by class initialization or function 
    according to the specified format.

    Parameters
    ----------
    obj : class, function
        The kind of class or function.
    locals : dict
        Local parameters in class.__init__ or function.
    format : str
        Parameters splicing format. 
        - log: [name \\n\\t param1 \\n\\t param2 \\n\\t ...]
        - rp: name(param1, param2, ...)

    Returns
    -------
    str : 
        A string specifying the format.

    Examples
    --------
    When you want to get the parameters of a class and its methods. Note: It is
    best to use `self` as the `obj` parameter during initialization and avoid 
    using the class to prevent class_name recognition errors of inherited classes.

    >>> class Base:
    ...     def __init__(self, param1, param2 = 2) -> None:
    ...         print(get_init_args(self, locals(), 'rp'))
    ...     def method(self, param3):
    ...         print(get_init_args(self.method, locals(), 'rp'))
    >>> class Sub(Base):
    ...     pass
    >>> obj_base = Base(1)
    >>> obj_base.method(3)
    >>> obj_sub = Sub(1)
    >>> obj_sub.method(3)
    Base(param1=1, param2=2)
    Base.method(param3=3)
    Sub(param1=1, param2=2)
    Sub.method(param3=3)

    Likewise, functions are supported.

    >>> def func(param1, param2 = 2):
    ...     print(get_init_args(func, locals(), 'rp'))
    >>> func(1)
    func(param1=1, param2=2)
    '''
    if inspect.isfunction(obj):
        obj_name = obj.__name__
    elif inspect.ismethod(obj):
        class_name = obj.__self__.__class__.__name__
        obj_name = f'{class_name}.{obj.__name__}'
    elif inspect.isclass(obj):
        obj_name = obj.__name__
    # for `self`
    elif isinstance(obj, object):
        obj_name = obj.__class__.__name__
        obj = obj.__init__
    else:
        raise TypeError(f'Unable to parse {obj}.')

    sig = inspect.signature(obj).parameters.keys()

    if format == 'log':
        s = f'[{obj_name}:\n'
        for key in sig:
            s += align_text(f'  [{key}]: ', str(locals[key])) + '\n'
        s += f']'
    elif format == 'rp':
        s = f'{obj_name}('
        kwargs = {key: locals[key] for key in sig}
        s = s + dict_to_str(kwargs) + ')'
    else:
        raise ValueError(f'{format} format is not supported.')

    return s


def find_mismatch_elements(key : Iterable[Any], query : Iterable[Any]):
    '''Check whether the values in the key are all in the query, and return 
    only those values that exist in key and not in query.

    Returns
    -------
    set
        List of elements in the key that do not exist in the query.
    
    Examples
    --------
    >>> list1, list2 = [0, 3, 4, 6], [1, 2, 3, 4, 5]
    >>> find_non_elements(list1, list2)
    [0, 6]
    '''
    mismatch_elements = [element for element in key if element not in query]
    return mismatch_elements


#!/usr/bin/env python
# coding: utf-8

"""
    @Desc    : A file handler to write in realtime.
    @Author  : SheepTAO
    @Time    : 2023-07-25
"""

import os
import csv


class Filer:
    def __init__(
        self, 
        path : str, 
        mode : str = 'a', 
        exist_ok : bool = False
    ) -> None:
        '''A simple file handler to write in realtime.

        Parameters
        ----------
        path : str
            File path. Support hierarchical folder structure.
        mode : str
            Mode of write.
        exist_ok : bool
            Whether to allow files that already exist.
        '''
        self.path = path
        self.mode = mode
        if os.path.exists(path) and not exist_ok:
            raise FileExistsError(f"File '{path}' already exists.")
        folder_path = os.path.dirname(path)
        os.makedirs(folder_path, exist_ok=True)

    def write(self, data : str) -> None:
        with open(self.path, self.mode) as f:
            f.write(data)


class CSVer(Filer):
    def __init__(self, path: str, mode: str = 'a') -> None:
        super().__init__(path, mode)

    def write(self, data: str) -> None:
        with open(self.path, self.mode) as f:
            w = csv.writer(f)
            w.writerows(data)



'''
References
----------
R. Mane et al., “FBCNet: A Multi-view Convolutional Neural Network for Brain-
Computer Interface,” arXiv.org. Accessed: Mar. 09, 2023. [Online]. 
Available: https://arxiv.org/abs/2104.01233v1
'''


import torch
from torch import nn

from .utils import Conv2dWithNorm, LinearWithNorm


class swish(nn.Module):
    def __init__(self):
        super(swish, self).__init__()

    def forward(self, x):
        return x * torch.sigmoid(x)


class LogVarLayer(nn.Module):
    def __init__(self, dim):
        super(LogVarLayer, self).__init__()
        self.dim = dim

    def forward(self, x):
        x = torch.clamp(x.var(dim=self.dim, keepdim=True), 1e-6, 1e6)
        return torch.log(x)


class FBCNet(nn.Module):
    def __init__(self, nCh, nTime, cls, bands=9, m=32, stride=4) -> None:
        super().__init__()
        self.stride = stride

        self.scb = nn.Sequential(
            Conv2dWithNorm(bands, m * bands, (nCh, 1), groups=bands, 
                           do_weight_norm=True, max_norm=2, padding=0),
            nn.BatchNorm2d(m * bands),
            swish(),
        )

        self.temporal_layer = LogVarLayer(dim=3)

        self.head = nn.Sequential(
            nn.Flatten(),
            LinearWithNorm(m*bands*stride, cls, do_weight_norm=True, max_norm=0.5),
            nn.LogSoftmax(dim=1)
        )

    def forward(self, x):
        x = self.scb(x)
        x = x.reshape([*x.shape[:2], self.stride, x.shape[3]//self.stride])
        x = self.temporal_layer(x)
        x = self.head(x)
        return x


if __name__ == '__main__':
    from torchinfo import summary
    net = FBCNet(22, 1000, 4).cuda()
    summary(net, (1, 9, 22, 1000))


#!/usr/bin/env python
# coding: utf-8

"""
    This module is used to load commonly used MI-EEG data.
    
    NOTE: All data is read during the initialization phase.
        - Defect: At the beginning, all the data needs to be read once, and the
        initialization time is longer. If the amount of data is huge, the 
        memory consumption is huge.
        - Advantage: In the process of multiple iterations, there is no need to
        read data repeatedly, and only one round of IO operation is required, 
        which saves time for multiple rounds of training.

    NOTE: If you want to build your own data set, you can refer to the sample 
        dataset to build it. Please reassign the `_event_id` and `_epochs` 
        attributes. In the final stage of initialization, please call the 
        `self.load_data()` to completely initialize your dataset.


    @Author  : SheepTAO
    @Time    : 2023-07-26
"""


import os
import mne
import numpy as np
from typing import Optional, List, Union, Literal
from scipy.io import loadmat

from .preprocessing import Preprocess, ComposePreprocess
from .transforms import Transforms, ComposeTransforms, SplitTrainTest
from ..utils import loger, verbose, DPEEG_SEED, get_init_args
from .functions import load


class EEGDataset:
    '''Base EEG dataset'''
    def _check_attr(self, attr : str) -> dict:
        '''Check if the attribute is None.'''
        if getattr(self, attr) is None:
            raise NotImplementedError(
                f"{self.__class__.__name__} model should implement '{attr}'.")
        return getattr(self, attr)

    @property
    def size(self) -> tuple:
        '''Return the current dimension of the dataset.'''
        sub = next(iter(self.dataset.values()))
        return (1, *sub['train'][0].shape[1:])

    @property
    def event_id(self) -> dict:
        '''Return the event label of the dataset and its corresponding id.'''
        return self._check_attr('_event_id')

    @property
    def cls_name(self) -> tuple:
        '''Return the event label of the dataset.'''
        return tuple(self.event_id.keys())

    @property
    def epochs(self) -> dict:
        '''Return the original Epochs data.'''
        return self._check_attr('_epochs')

    @property
    def dataset(self) -> dict:
        '''Return the loaded data.'''
        return self._check_attr('_dataset')

    def __init__(
        self,
        preprocess : Optional[Preprocess] = None,
        transforms : Optional[Transforms] = None,
        test_size : float = .25, 
        seed : int = DPEEG_SEED,
        verbose : Optional[Union[str, int]] = None, 
    ) -> None:
        '''EEG Dataset abstract base class.

        Parameters
        ----------
        preprocess : Preprocess, optional
            Apply preprocessing on epochs.
        transforms : Transforms, optional
            Apply pre-transforms on dataset.
        test_size : float
            Split the training set and test set proportions. If the dataset is
            already split, it will be ignored.
        seed : int
            Random seed when splitting.
        verbose : int, str, optional
            Log level of mne.
        '''
        mne.set_log_level(verbose)

        self._repr = None
        self._preprocess = preprocess
        self._transforms = transforms
        self._test_size = test_size
        self._seed = seed
        self._dataset = None
        self._verbose = verbose

        # NOTE
        # Please make sure the following attributes are correctly overridden
        self._event_id = None        # task name and its corresponding label
        self._epochs = None         # each subject and its corresponding Epochs
    
    def load_data(
        self, 
        split : bool = False,
        unit_factor : float = 1e6,
        include_tmax : bool = False,
    ) -> None:
        '''Extract data from Epochs and split.

        Parameters
        ----------
        split : bool
            Whether `self.raw` has been splited.
        unit_factor : float
            Unit factor to convert the units of uv to v.
        include_tmax : bool
            Whether to include the last timestamp is used to solve the data 
            timestamp problem after resampling and will be removed in future
            versions.

        Notes
        -----
        Avoid data leakage when you split data.
        '''
        if self._preprocess:
            pres = ComposePreprocess(self._preprocess)
            self._epochs = pres(self.epochs)

        dataset = {}
        for sub, sEpochs in self.epochs.items():
            if not split:
                data = sEpochs.crop(include_tmax=include_tmax).get_data()
                label = sEpochs.events[:, -1]
                dataset[sub] = [data * unit_factor, label]
            else:
                dataset[sub] = {}
                for mode, mEpochs in sEpochs.items():
                    data = mEpochs.crop(include_tmax=include_tmax).get_data()
                    label = mEpochs.events[:, -1]
                    dataset[sub][mode] = [data * unit_factor, label]

        # split the dataset before transforms
        if self._transforms:
            trans = ComposeTransforms(self._transforms)
            if not split:
                trans.insert(0, SplitTrainTest(self._test_size, self._seed))
            self._dataset = trans(dataset, verbose=self._verbose)
        else:
            self._dataset = dataset
        loger.info('[Loading dataset done]')

    def process_interval(
        self,
        tmin : float,
        tmax : float,
        interval : Optional[tuple] = None,
        baseline : Optional[tuple] = None,
    ) -> tuple:
        '''Preprocessing of data extraction time.

        This function is used to correct the extraction timestamps and baseline
        adjustments of different data sets according to the interval. Can be 
        called as needed.

        Parameters
        ----------
        tmin : float
             Start time (in second) of the epoch, relative to the dataset spec-
             ific task interval e.g. tmin = 1 would mean the epoch will start 1
             second after the beginning of the task as defined by the dataset.
        tmax : float, optional
            End time (in second) of the epoch, relative to the beginning of the
            dataset specific task interval. tmax = 5 would mean the epoch will 
            end 5 second after the beginning of the task as defined in the data
            set.
        interval : tuple of length 2, optional
            Imagery interval as defined in the dataset description.
        baseline : tuple of length 2, optional
            The time interval to consider as “baseline” when applying baseline 
            correction. If None, do not apply baseline correction. If a tuple 
            (a, b), the interval is between a and b (in seconds), including the
            endpoints.

        Returns
        -------
        If baseline is not None, will return the corrected `tmin`, `tmax` and 
        `baseline`, or only return the corrected `tmin`, `tmax` and `None`.
        '''
        if tmin >= tmax:
            raise ValueError('tmax must be greater than tmin')

        if interval:
            if len(interval) == 2:
                if interval[0] >= interval[1]:
                    raise ValueError('End time of the interval must be greater'
                                     ' than the start time.')
                tmin, tmax = interval[0] + tmin, interval[0] + tmax

                if baseline:
                    baseline = (
                        baseline[0] + interval[0], baseline[1] + interval[0]
                    )
                    bmin = baseline[0] if baseline[0] < tmin else tmin
                    bmax = baseline[1] if baseline[1] > tmax else tmax
                    return tmin, tmax, (bmin, bmax)
                else:
                    return tmin, tmax, baseline

            else:
                raise ValueError('Interval should be the start and end time '
                                 f'points, but got {len(interval)} numbers.')
        else:
            return tmin, tmax, baseline

    def __getitem__(self, args) -> dict:
        '''Return the index position data of the corresponding subject
        and the corresponding set.

        Parameters
        ----------
        args : sub, index (optional)
            sub : int
                Number of subject.
            index : int, slice, optional
                The data index corresponding to the training set. If index is 
                None, all the data (train and test) of the subject will be re-
                turned as a dictionary.

        Returns
        -------
        dict - A dict contain train set (and test set).

        Example
        -------
            >>> dataset[1]
            {'train': (sub1[data], sub1[label]), 
             'test': (sub1[data], sub1[label])}
            >>> dataset[2, :2]
            (sub2['train'][data[:2]], sub2['train'][label[:2]])
        '''
        if isinstance(args, int):
            sub = args
            index = None
        else:
            num = len(args)
            if num != 2:
                raise ValueError(f'Index must be 1 to 2 numbers, but got {num}.')
            else:
                sub, index = args[0], args[1]

        if index is not None:
            return {
                'train' : (self.dataset[sub]['train'][0][index],
                           self.dataset[sub]['train'][1][index])
            }
        else:
            return {
                'train': tuple(self.dataset[sub]['train']),
                'test': tuple(self.dataset[sub]['test'])
            }

    def __repr__(self) -> str:
        if self._repr:
            return self._repr
        else:
            class_name = self.__class__.__name__
            return f'{class_name} not implement attribute `self._repr`.'

    def items(self):
        return self.dataset.items()

    def keys(self):
        return self.dataset.keys()

    def values(self):
        return self.dataset.values()


class PhysioNet(EEGDataset):
    '''Physionet MI Dataset.
    channels=64; subjects=1-109[88, 92, 100, 104], tasks={0:left hand, 1:right 
    hand, 2:hands, 3:feet}; duration=3s; freq=160Hz, sessions=1.
    '''
    @verbose
    def __init__(
        self,
        subjects : Optional[List[int]] = None,
        tmin : float = 0,
        tmax : float = 1,
        preprocess : Optional[Preprocess] = None,
        transforms : Optional[Transforms] = None,
        test_size : float = .25,
        picks : Optional[List[str]] = None,
        baseline = None,
        seed : int = DPEEG_SEED,
        verbose : Optional[str] = None
    ) -> None:
        '''Physionet MI Dataset.

        Parameters
        ----------
        subjects : list of int, optional
            List of subject number. If None, all subjects will be loaded. 
        tmin, tmax : float
            Start and end time of the epochs in seconds, relative to the time-
            locked event. The closest or matching samples corresponding to the
            start and end time are included. Default is start and end time of 
            epochs, respectively.
        picks : list of str, optional
            Channels to include. If None, pick all channels.
        preprocess : Preprocess, optional
            Apply preprocessing on epochs.
        transforms : Transforms, optional
            Apply pre-transforms on dataset.
        test_size : float
            Split the training set and test set proportions. If the dataset is
            already split, it will be ignored.
        seed : int
            Random seed when splitting.
        '''
        super().__init__(preprocess, transforms, test_size, seed, verbose)
        self._repr = get_init_args(self, locals())
        loger.info('Reading PhysionetMI Dataset ...')

        from moabb.datasets import PhysionetMI
        dataset = PhysionetMI()
        tmin, tmax, baseline = \
            self.process_interval(tmin, tmax, dataset.interval, baseline)
        self._base_raw = dataset.get_data(subjects)
        self.bad_sub = [88, 92, 100, 104]
        self._event_id = {
            'left_hand': 0,
            'right_hand': 1,
            'hands': 2,
            'feet': 3
        }

        self._epochs = {}
        for sub, sess in self._base_raw.copy().items():
            loger.info(f'\nCreate sub{sub} Epochs ...')
            if sub in self.bad_sub:
                continue
            epochs_ses = []
            for raw in sess['session_0'].values():

                ann = mne.events_from_annotations(raw, regexp='^(?!rest$).*$')
                # update events (labels)
                ann[0][:, -1] = ann[0][:, -1]-1 if 'left_hand' in ann[1].keys() \
                    else ann[0][:, -1]+1
                # update event_id
                for key in ann[1].keys():
                    ann[1][key] = self._event_id[key]

                events = ann[0][:-1]
                epochs_ses.append(
                    mne.Epochs(raw, events, ann[1], tmin, tmax, baseline, picks, 
                               preload=True)
                )

            self._epochs[sub] = mne.concatenate_epochs(epochs_ses, verbose='ERROR')

        self.load_data()


class BCICIV2A(EEGDataset):
    '''BCI Competition IV Data sets 2a.
    channels=22; subjects=1-9; tasks={0:left hand, 1:right hand, 2:feet, 3:tongue};
    duration=4s; freq=250Hz; sessions=2.
    '''
    @verbose
    def __init__(
        self,
        subjects : Optional[List[int]] = None,
        tmin : float = 0,
        tmax : float = 4,
        preprocess : Optional[Preprocess] = None,
        transforms : Optional[Transforms] = None,
        test_size : float = .25,
        mode : Literal['single_ses', 'cross_ses', 'mixed_ses'] = 'cross_ses',
        picks : Optional[List[str]] = None,
        baseline = None,
        seed : int = DPEEG_SEED,
        verbose : Optional[str] = None
    ) -> None:
        '''BCI Competition IV Data sets 2a.

        Parameters
        ----------
        subjects : list of int, optional
            List of subject number. If None, all subjects will be loaded. 
        tmin, tmax : float
            Start and end time of the epochs in seconds, relative to the time-
            locked event. The closest or matching samples corresponding to the
            start and end time are included. Default is start and end time of 
            epochs, respectively.
        picks : list of str, optional
            Channels to include. If None, pick all channels.
        preprocess : Preprocess, optional
            Apply preprocessing on epochs.
        transforms : Transforms, optional
            Apply pre-transforms on dataset.
        mode : str, optional
            If mode = 'single_ses', training data and test data will only use
            session 1. If mode = 'cross_ses', training data and test data will
            use session 1 and 2, respectively. If mode = 'mixed_ses', training
            data and test data will use both session 1 and 2.
        test_size : float
            Split the training set and test set proportions. If the dataset is
            already split, it will be ignored.
        seed : int
            Random seed when splitting.
        '''
        super().__init__(preprocess, transforms, test_size, seed, verbose)
        self._repr = get_init_args(self, locals())
        loger.info('Reading BCICIV 2A Dataset ...')

        from moabb.datasets import BNCI2014001
        dataset = BNCI2014001()
        tmin, tmax, baseline = \
            self.process_interval(tmin, tmax, dataset.interval, baseline)
        self._base_raw = dataset.get_data(subjects)
        self._event_id = {
            'left_hand': 0,
            'right_hand': 1,
            'feet': 2,
            'tongue': 3
        }

        self._epochs = {}
        for sub, sessions in self._base_raw.copy().items():
            loger.info(f'Create sub{sub} Epochs ...')
            epochs_ses_one, epochs_ses_two = [], []
            for session, runs in sessions.items():
                for raw in runs.values():

                    events = mne.find_events(raw, 'stim')
                    events[:, -1] -= 1
                    epochs = mne.Epochs(raw, events, self._event_id, tmin, 
                                        tmax, baseline, picks, preload=True)
                    epochs.drop_channels(['stim', 'EOG1', 'EOG2', 'EOG3'])

                    if session == 'session_T':
                        epochs_ses_one.append(epochs)
                    else:
                        epochs_ses_two.append(epochs)

            if mode == 'single_ses':
                self._epochs[sub] = mne.concatenate_epochs(epochs_ses_one)
            elif mode == 'cross_ses':
                self._epochs[sub] = {}
                self._epochs[sub]['train'] = mne.concatenate_epochs(epochs_ses_one)
                self._epochs[sub]['test'] = mne.concatenate_epochs(epochs_ses_two)
            elif mode == 'mixed_ses':
                self._epochs[sub] = mne.concatenate_epochs(
                    epochs_ses_one + epochs_ses_two
                )
            else:
                raise ValueError(f'Mode does not support {mode}.')

        split = True if mode == 'cross_ses' else False
        self.load_data(split)


class BCICIV2B(EEGDataset):
    '''BCI Competition IV Data sets 2b.
    channels=3; subjects=1-9; tasks={0:left hand, 1:right hand}; duration=4s; 
    freq=250Hz; sessions=5.
    '''
    @verbose
    def __init__(
        self,
        subjects : Optional[List[int]] = None,
        tmin : float = 0,
        tmax : float = 4,
        preprocess : Optional[Preprocess] = None,
        transforms : Optional[Transforms] = None,
        test_size : float = .25,
        mode : Literal['single_ses', 'cross_ses', 'mixed_ses'] = 'cross_ses',
        test_sessions : Optional[List[int]] = None,
        picks : Optional[List[str]] = None,
        baseline = None,
        seed : int = DPEEG_SEED,
        verbose : Optional[str] = None,
    ) -> None:
        '''BCI Competition IV Data sets 2b.

        Parameters
        ----------
        subjects : list of int, optional
            List of subject number. If None, all subjects will be loaded. 
        tmin, tmax : float
            Start and end time of the epochs in seconds, relative to the time-
            locked event. The closest or matching samples corresponding to the
            start and end time are included. Default is start and end time of 
            epochs, respectively.
        picks : list of str, optional
            Channels to include. If None, pick all channels. Default is None.
        preprocess : Preprocess, optional
            Apply preprocessing on epochs.
        transforms : Transforms, optional
            Apply pre-transforms on dataset.
        mode : str, optional
            If mode = 'cross_ses', training data and test data will use differ-
            ent sessions respectively. At this time, you need to specify `test_
            ses` to select which sessions are used as the test set. If mode = 
            'mixed_ses', training data and test data will use all sessions. If 
            mode='single_ses', the training and test sets will come from the 
            same session, and `test_ses` specifies the session used.
        test_sessions : list, optional
            Specify the selected session as the test set.
        test_size : float
            Split the training set and test set proportions. If the dataset is
            already split, it will be ignored.
        seed : int
            Random seed when splitting.
        '''
        super().__init__(preprocess, transforms, test_size, seed, verbose)
        self._repr = get_init_args(self, locals())
        loger.info('Reading BCICIV 2B Dataset ...')

        from moabb.datasets import BNCI2014004
        dataset = BNCI2014004()
        tmin, tmax, baseline = \
            self.process_interval(tmin, tmax, dataset.interval, baseline)
        self._base_raw = dataset.get_data(subjects)
        self._event_id ={'left_hand': 0, 'right_hand': 1}

        test_ses = None
        if test_sessions:
            test_ses = [f'session_{ses}' for ses in test_sessions]

        self._epochs = {}
        for sub, sessions in self._base_raw.copy().items():
            loger.info(f'Create sub{sub} Epochs ...')
            epochs_ses_train, epochs_ses_test = [], []
            for session, runs in sessions.items():
                for run, raw in runs.items():

                    events = mne.find_events(raw)
                    events[:, -1] -= 1

                    epochs = mne.Epochs(raw, events, self._event_id, tmin, 
                                        tmax, baseline, picks, preload=True)
                    epochs.drop_channels(['stim', 'EOG1', 'EOG2', 'EOG3'])
                    print(f'sub{sub}-{session}-{run}: {len(events)}')

                    if mode in ['single_ses', 'cross_ses']:
                        assert test_ses, f'In {mode} mode, test_sessions cannot be None.'
                        if session in test_ses:
                            epochs_ses_test.append(epochs)
                        else:
                            epochs_ses_train.append(epochs)

            if mode == 'single_ses':
                self._epochs[sub] = epochs_ses_test[0]
            elif mode == 'cross_ses':
                self._epochs[sub] = {
                    'train': mne.concatenate_epochs(epochs_ses_train),
                    'test': mne.concatenate_epochs(epochs_ses_test)
                }
            elif mode == 'mixed_ses':
                self._epochs[sub] = mne.concatenate_epochs(
                    epochs_ses_test + epochs_ses_train
                )
            else:
                raise ValueError(f'Mode does not support {mode}.')

        split = True if mode == 'cross_ses' else False
        self.load_data(split=split)


class HGD(EEGDataset):
    '''High Gamma Dataset.
    channels=128; subjects=1-14; tasks={0:feet, 1:left hand, 2:rest, 3:right hand};
    duration=4s; freq=500Hz; sessions=1.

    The dataset per subject included approximately 1040 trials over 13 runs. 
    The first 11 runs (approximately 880 trials) were used for the training 
    and the last 2 runs (approximately 160 trials) were used for evaluation.
    '''
    @verbose
    def __init__(
      self,
      subjects : Optional[List[int]] = None,
      tmin : float = 0,
      tmax : float = 4,
      preprocess : Optional[Preprocess] = None,
      transforms : Optional[Transforms] = None,
      test_size : float = .25,
      picks : Optional[List[str]] = None,
      baseline = None,
      seed : int = DPEEG_SEED,
      verbose : Optional[str] = None,
    ) -> None:
        '''High Gamma Dataset.

        Parameters
        ----------
        subjects : list of int, optional
            List of subject number. If None, all subjects will be loaded. 
        tmin, tmax : float
            Start and end time of the epochs in seconds, relative to the time-
            locked event. The closest or matching samples corresponding to the
            start and end time are included. Default is start and end time of 
            epochs, respectively.
        picks : list of str, optional
            Channels to include. If None, pick all channels.
        preprocess : Preprocess, optional
            Apply preprocessing on epochs.
        transforms : Transforms, optional
            Apply pre-transforms on dataset.
        test_size : float
            Split the training set and test set proportions. If the dataset is
            already split, it will be ignored.
        seed : int
            Random seed when splitting.
        '''
        super().__init__(preprocess, transforms, test_size, seed, verbose)
        self._repr = get_init_args(self, locals())
        loger.info('Reading High Gamma Dataset ...')

        from moabb.datasets import Schirrmeister2017
        dataset = Schirrmeister2017()
        tmin, tmax, baseline = \
            self.process_interval(tmin, tmax, dataset.interval, baseline)
        self._base_raw = dataset.get_data(subjects)
        self._event_id = {
            'feet': 0, 
            'left_hand': 1, 
            'rest': 2, 
            'right_hand': 3
        }

        self._epochs = {}
        for sub, sess in self._base_raw.items():
            loger.info(f'Create sub{sub} Epochs ...')
            for mode, run in sess['session_0'].items():
                events, _ = mne.events_from_annotations(run)
                # update events
                events[:, -1] -= 1
                epochs = mne.Epochs(run, events, self._event_id, tmin, tmax,
                                    baseline, picks, preload=True)
                self._epochs.setdefault(sub, {})[mode] = epochs

        self.load_data(split=True, include_tmax=True)
        

class SEED(EEGDataset):
    @verbose
    def __init__(
      self,
      path : str,
      session : int = 0,
      transforms : Optional[Transforms] = None,
      test_size : float = .25,
      seed : int = DPEEG_SEED,
      verbose : Optional[str] = None,
    ) -> None:
        super().__init__(None, transforms, test_size, seed, verbose)
        self._repr = get_init_args(self, locals())
        loger.info('Reading SEED Dataset ...')

        self._event_id = {
            'negative': 0,
            'netural': 1,
            'positive': 2
        }

        subjects_list = os.listdir(os.path.abspath(path))
        subjects_list.remove('label.mat')
        subjects_list.remove('readme.txt')
        subjects_list.sort(key=lambda x : int(x.split('_')[0]))

        sub_session = [subjects_list[s] for s in range(session, 45, 3)]
        sub_session_path = [os.path.join(path, p) for p in sub_session]

        label_path = os.path.join(path, 'label')
        label = loadmat(label_path)
        label = label['label'].squeeze(0) + 1

        dataset = {}
        for sub, data_path in enumerate(sub_session_path):
            all_trial = loadmat(data_path)
            all_trial.pop('__header__')
            all_trial.pop('__version__')
            all_trial.pop('__globals__')

            trial_list = []
            for trial in all_trial.values():
                trial_list.append(np.expand_dims(trial[:, :36000], 0))
            data = np.concatenate(trial_list)
            dataset[sub+1] = [data, label]

        if transforms:
            trans = ComposeTransforms(transforms)
            trans.insert(0, SplitTrainTest(test_size, seed))
        else:
            trans = SplitTrainTest(test_size, seed)
        self._dataset = trans(dataset, verbose=verbose)


#!/usr/bin/env python
# coding: utf-8

"""
    This module defines with what experiments the module is trained and validated.

    experiments:
        1 - kFoldCV : K-Fold cross-validation.
        2 - HoldOut : hold out validation.

    @Author  : SheepTAO
    @Time    : 2023-07-26
"""


import torch
import os, abc
from torch import Tensor
from typing import Union, Optional, Tuple, Literal
from sklearn.model_selection import StratifiedKFold
from torchmetrics.functional.classification.cohen_kappa import cohen_kappa
from torchmetrics.aggregation import MeanMetric, CatMetric

import dpeeg
from .classifier import TrainClassifier
from ..data.datasets import EEGDataset
from ..data.functions import split_train_test, merge_train_test, check_dataset, check_sub_data
from ..data.transforms import Transforms
from ..tools import Timer, Filer, Logger
from ..utils import DPEEG_SEED, DPEEG_DIR, get_init_args
from .evaluate import save_cm_img


class Experiment(abc.ABC):
    '''Experiment base class.
    '''
    def __init__(
        self,
        trainer : TrainClassifier,
        out_folder : Optional[str] = None,
        verbose : Union[int, str] = 'INFO'
    ) -> None:
        '''Initialize the basic parameters of the experiment.

        Parameters
        ----------
        trainer : Train
            Trainer used for training module on dataset.
        out_folder : str, optional
            Store all experimental results in a folder named with the model 
            class name in the specified folder. Default is 
            '~/dpeeg/out/exp_name/model_name/'.
        verbose : int, str
            The log level of console. Default is INFO. Mainly used for debugg-
            ing.

        Notes
        -----
        The training results of all models for each subject will be saved under
        the out_folder directory.
        '''
        self._repr = None
        self.trainer = trainer
        self.verbose = verbose

        # create loger and timer
        self.loger = Logger('dpeeg_exp', clevel=verbose)
        self.timer = Timer()

        # set output folder
        net = trainer.net.__class__.__name__
        exp = self.__class__.__name__
        self.out_folder = os.path.join(os.path.abspath(out_folder), net, exp) \
            if out_folder else os.path.join(DPEEG_DIR, 'out', net, exp)
        os.makedirs(self.out_folder, exist_ok=True)
        self.loger.info(f'Results will be saved in folder: {self.out_folder}')

    @abc.abstractmethod
    def _run_sub(
        self,
        trainset : Union[tuple, list],
        testset : Union[tuple, list],
        cls_name : Union[tuple, list],
        sub : str,
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, dict]:
        '''Train a model on the specified subject data.

        This function will be called by `run` function to conduct experiments
        on the data of each individual subject. Reconstruct the model training 
        process according to different experimental requirements.

        Parameters
        ----------
        trainset : tuple, list
            Dataset used for training module. If type is tuple or list, dataset
            should be (data, labels).
        testset : tuple, list
            Dataset used for evaluating moduel. If type is tuple or list, data-
            set should be (data, labels).
        sub : str, optional
            Create a subdirectory of out_folder to store all yield results dur-
            ing subject training.
        cls_name : tuple, list
            The name of dataset labels.

        Returns
        -------
        Return test_acc, test_kappa, test_preds, test_target and results.
        '''
        pass

    def _trans_dataset(
        self, 
        trainset : Union[list, tuple], 
        testset : Union[list, tuple]
    ) -> tuple:
        '''Apply pre-transforms on dataset.

        Generate a new virtual subject based on the input trainset and testset,
        then transform the new dataset and finally return the new trainset and
        testset.

        Parameters
        ----------
        trainset, testset : list, tuple
            List of data and labels.

        Returns
        -------
        Transformed trainset and testset.
        '''
        if self.transforms is not None:
            sub_dataset = {-1 : {'train':list(trainset), 'test':list(testset)}}
            sub_dataset = self.transforms(sub_dataset)
            trainset = sub_dataset[-1]['train']
            testset = sub_dataset[-1]['test']

        return trainset, testset
        
    def _process_sub_dataset(self, sub : int) -> tuple:
        '''Preprocess each subject's dataset.

        Different preprocessing operations are performed on the dataset accord-
        ing to different experimental requirement. By default, the training set
        and test set for each subject are returned.

        Parameters
        ----------
        sub : int
            Subjects currently undergoing the experiment.

        Returns
        -------
        trainset : tuple
            Tuple containing training data and labels.
        testset : tuple
            Tuple containing test data and labels.
        '''
        return self._trans_dataset(
            self.dataset[sub]['train'], self.dataset[sub]['test']
        )

    def run(
        self,
        dataset : Union[EEGDataset, dict],
        cls_name : Union[tuple, list],
        dataset_name : Optional[str] = None,
        transforms : Optional[Transforms] = None,
        desc : Optional[str] = None,
    ) -> dict:
        '''Train models separately for each subject.

        This function will internally call the `_run_sub` function for each 
        subject, and save the final results together.

        Parameters
        ----------
        dataset : EEGDataset, dict
            Dataset structured as:\n
            {0 : { 'train': (array(nsample x channels x times), array(nsample)),
                   'test': (array(nsample x channels x times), array(nsample))},
             1 : {...},
            ...}\n
            where array can be `numpy.ndarray` or `torch.Tensor`. Please make 
            sure that the dataset has been split.
        cls_name : tuple, list
            The name of dataset labels.
        dataset_name : str, optional
            The dataset name to use. If None, `dataset.__class__.__name__` will 
            be used as the folder to save experimental results. If the type of 
            the dataset is dict, Please provide a dataset name to prevent 
            confusing results.
        transforms : Transforms, optional
            Apply pre-transforms on dataset. Transformations will be apply 
            during the experiment on each subject's dataset. The rationable
            behind this method lies in deferring data manipulation, especially
            for certain transformations that could potentially enlarge the 
            dataset's memory footprint. This delay allows for the execution of 
            data manipulation after subject-independent experiment have concat-
            enated the relevant data (Time for Space) or the experiment subject
            are ready, mitigating the risk of memory overflow.
        desc : str, optional
            Add a short description to the current experiment.

        Returns
        -------
        Return a dict of all subjects and corresponding experimental results.
        '''
        if dataset_name:
            self.data_folder = os.path.join(self.out_folder, dataset_name)
        elif isinstance(dataset, EEGDataset):
            self.data_folder = \
                os.path.join(self.out_folder, dataset.__class__.__name__)
        else:
            self.data_folder = os.path.join(self.out_folder, 'Custom Dataset')
        os.makedirs(self.data_folder)

        filer = Filer(os.path.join(self.data_folder, 'summary.txt'))
        filer.write(f'[Start Time]: {self.timer.ctime()}\n')
        filer.write(f'[DPEEG Version]: {dpeeg.__version__}\n')
        filer.write(f'[Description]: {desc}\n')
        filer.write(str(self) + '\n')
        if isinstance(dataset, EEGDataset):
            filer.write(str(dataset) + '\n')
            dataset = dataset.dataset
        else:
            filer.write('[Custom dataset]\n')

        self.dataset = dataset
        self.transforms = transforms

        # save all sub results
        results = {}
        acc_metric, kappa_metric = MeanMetric(), MeanMetric()
        preds_metric, target_metric = CatMetric(), CatMetric()

        # update root timer and start cross validation for each subject
        self.timer.start()
        self.loger.info('=' * 50)
        for sub in dataset.keys():
            self.loger.info(f'\n[Subject-{sub} Training ...]')
            trainset, testset = self._process_sub_dataset(sub)
            self.loger.info('-' * 50)

            # run one cross validation
            test_acc, test_kappa, test_preds, test_target, sub_exp_res = \
                self._run_sub(trainset, testset, cls_name, f'sub{sub}')

            results[f'sub_{sub}'] = sub_exp_res
            acc_metric.update(test_acc)
            kappa_metric.update(test_kappa)
            preds_metric.update(test_preds)
            target_metric.update(test_target)

            filer.write(f'---------- Sub_{sub} ----------\n')
            filer.write(f'Acc = {test_acc*100:.2f}% | Kappa = '
                        f'{test_kappa:.2f}\n\n')

        acc = acc_metric.compute()
        kappa = kappa_metric.compute()

        filer.write(f'---------- MODEL ----------\n')
        filer.write(f'Acc = {acc*100:.2f}% | Kappa = {kappa:.2f}\n')
        save_cm_img(preds_metric.compute(), target_metric.compute(), 
                    cls_name, os.path.join(self.data_folder, 'CM.png'))
        torch.save(results, os.path.join(self.data_folder, f'results.pt'))

        h, m, s = self.timer.stop()
        self.loger.info(f'\n[All subjects finished]')
        self.loger.info(f'Cost time = {h}H:{m}M:{s:.2f}S')
        self.loger.info('=' * 50)
        self.loger.info(f'[Acc = {acc*100:.2f}% | Kappa = {kappa:.2f}]')

        return results

    def __repr__(self) -> str:
        if self._repr:
            return self._repr
        else:
            class_name = self.__class__.__name__
            return f'{class_name} not implement attribute `self._repr`.'


class KFold(Experiment):
    def __init__(
        self,
        trainer : TrainClassifier,
        k : int = 5,
        out_folder : Optional[str] = None,
        max_epochs_s1 : int = 1500,
        max_epochs_s2 : int = 600,
        no_increase_epochs : int = 100,
        second_stage : bool = True,
        load_best_state : bool = True,
        var_check : Literal['val_inacc', 'val_loss'] = 'val_inacc',
        isolate_testset : bool = True,
        shuffle : bool = True,
        seed : int = DPEEG_SEED,
        verbose : Union[int, str] = 'INFO',
    ) -> None:
        '''K-Fold cross validation experiment.

        Parameters
        ----------
        trainer : Train
            Trainer used for training module on dataset.
        k : int, optional
            k of k-Fold.
        out_folder : str, optional
            Store all experimental results in a folder named with the model 
            class name in the specified folder. Default is 
            '~/dpeeg/out/model_name/'.
        max_epochs_s1, max_epochs_s2 : int
            Maximum number of epochs in the x stage of training. Default is
            1500 and 600 respectively.
        no_increase_epochs : int
            Maximum number of consecutive epochs when the accuracy of the first
            stage validation set has no relative improvement.
        second_stage : bool
            If True, two-stage training will be performed.
        load_best_state : bool
            If True, stage 2 will retrain based on the best state in stage 1.
        var_check : str
            The best value ('val_inacc'/'val_loss') to check while determining 
            the best model which will be used for parameter initialization in
            the second stage of model training.
        isolate_testset : bool
            By default, the test set is independent, that is, the k-fold cross-
            validation at this time only divides the training set and the 
            verification set based on the training set to implement an early
            stopping mechanism, and finally evaluates on the isolated test set.
            If False, the test set is for each fold of k-fold cross-validation.
        shuffle : bool
            Shuffle before kfold.
        seed : int
            Seed of random for review.

        Notes
        -----
        If `isolate_testset` False, please provide the `transforms` parameter
        of the `run` function to avoid data leakage caused by operations such 
        as data augmentation in advance.

        TODO: Nested Cross-Validation.
        '''
        super().__init__(trainer, out_folder, verbose)

        self._repr = get_init_args(self, locals())
        self.max_epochs_s1 = max_epochs_s1
        self.max_epochs_s2 = max_epochs_s2
        self.no_increase_epochs = no_increase_epochs
        self.second_stage = second_stage
        self.load_best_state = load_best_state
        self.var_check = var_check
        self.isolate_testset = isolate_testset

        # create a stratified k-fold index
        self.skf = StratifiedKFold(k, shuffle=shuffle, random_state=seed)

    def _process_sub_dataset(self, sub: int) -> tuple:
        '''Whether to merge subject data depends on the experimental mode.
        '''
        if self.isolate_testset:
            return super()._process_sub_dataset(sub)

        if check_sub_data(sub, self.dataset[sub]):
            trainset = merge_train_test(*self.dataset[sub].values())
        else:
            trainset = self.dataset[sub]

        return trainset, None

    def _run_sub(
        self,
        trainset : Union[tuple, list],
        testset : Union[tuple, list],
        cls_name : Union[tuple, list],
        sub : str,
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, dict]:
        '''Basic K-Fold cross validation function.

        Returns
        -------
        Return test_acc, test_kappa, test_preds, test_target and results :\n
        results = {
            'expNo_1' : { ... },
            'expNo_2' : { ... },
            .
            .
        }
        '''
        # save all exp results
        results = {}
        train_acc_metric, val_acc_metric, test_acc_metric = \
            MeanMetric(), MeanMetric(), MeanMetric()
        test_preds_metric, test_target_metric = CatMetric(), CatMetric()

        # set subject's reults storage path
        sub_folder = os.path.join(self.data_folder, sub)
        os.makedirs(sub_folder)

        # store experiment results
        filer = Filer(os.path.join(sub_folder, 'summary.txt'))
        self.timer.start('kfold')

        # run k-fold
        for idx, (trainIdx, valIdx) in enumerate(self.skf.split(*trainset)):
            self.loger.info(f'\n# ------ {sub} - expNo.{idx+1} ------ #')

            if self.isolate_testset:
                train_set = (trainset[0][trainIdx], trainset[1][trainIdx])
                val_set = (trainset[0][valIdx], trainset[1][valIdx])
                test_set = testset
            else:
                train_set = [trainset[0][trainIdx], trainset[1][trainIdx]]
                test_set = [trainset[0][valIdx], trainset[1][valIdx]]
                train_set, test_set = self._trans_dataset(train_set, test_set)

                train_x, val_x, train_y, val_y = split_train_test(*train_set)
                train_set = (train_x, train_y)
                val_set = (val_x, val_y)

            # train one fold
            exp_path = os.path.join(sub_folder, f'expNo_{idx + 1}')
            res = self.trainer.fit_with_val(
                trainset=train_set,
                valset=val_set,
                testset=test_set,
                log_dir=exp_path,
                cls_name=cls_name,
                max_epochs_s1=self.max_epochs_s1,
                max_epochs_s2=self.max_epochs_s2,
                no_increase_epochs=self.no_increase_epochs,
                second_stage=self.second_stage,
                load_best_state=self.load_best_state,
                var_check=self.var_check # type: ignore
            )

            # save all results
            results[f'expNo_{idx}'] = res
            train_res, val_res, test_res = res['train'], res['val'], res['test']
            train_acc_metric.update(train_res['acc'])
            val_acc_metric.update(val_res['acc'])
            test_acc_metric.update(test_res['acc'])
            test_preds_metric.update(test_res['preds'])
            test_target_metric.update(test_res['target'])

            # save the confusion matirx image
            save_cm_img(test_res['preds'], test_res['target'], cls_name,
                        os.path.join(exp_path, f'{sub}_expNo_{idx+1}_CM.png'))

            filer.write(f'---------------- expNo_{idx+1} ----------------\n')
            filer.write(f'Acc : Train={train_res["acc"]:.4f}' +
                        f'|Val={val_res["acc"]:.4f}' +
                        f'|Test={test_res["acc"]:.4f}\n\n')

        h, m, s = self.timer.stop('kfold')
        self.loger.info(f'\n[Cross-Validation Finish]')
        self.loger.info(f'Cost time = {h}H:{m}M:{s:.2f}S')

        train_acc = train_acc_metric.compute()
        val_acc = val_acc_metric.compute()
        test_acc = test_acc_metric.compute()
        self.loger.info(f'Acc : train={train_acc:.4f} | val={val_acc:.4f} | ' +
                        f'test={test_acc:.4f}')

        test_preds = test_preds_metric.compute()
        test_target = test_target_metric.compute()
        save_cm_img(test_preds, test_target, cls_name,
                    os.path.join(sub_folder, f'{sub}_CM.png'))

        test_kappa = cohen_kappa(
            test_preds_metric.compute(), test_target_metric.compute(),
            task = 'multiclass', num_classes = len(cls_name)
        )

        # store the subject accuracy and cohen kappa on test set
        filer.write(f'--------- TestAvg ---------\n')
        filer.write(f'Acc = {test_acc*100:.2f}% | Kappa = {test_kappa:.2f}\n')

        return test_acc, test_kappa, test_preds, test_target, results


class Holdout(Experiment):
    def __init__(
        self,
        trainer : TrainClassifier,
        out_folder : Optional[str] = None,
        max_epochs_s1 : int = 1500,
        no_increase_epochs : int = 200,
        var_check : Optional[
            Literal['train_inacc', 'train_loss', 'val_inacc', 'val_loss']
        ] = None,
        split_val : bool = True,
        test_size : float = 0.25,
        second_stage : bool = True,
        load_best_state : bool = True,
        max_epochs_s2 : int = 600,
        seed : int = DPEEG_SEED,
        verbose : Union[int, str] = 'INFO'
    ) -> None:
        '''Holdout cross validation experiment.

        Validate the model using holdout cross validation, supports two model 
        training methods - only training set and training set split into valid-
        ation set.

        Parameters
        ----------
        trainer : Train
            Trainer used for training module on dataset.
        out_folder : str, optional
            Store all experimental results in a folder named with the model 
            class name in the specified folder. Default is 
            '~/dpeeg/out/model_name/'.
        max_epochs_s1, max_epochs_s2 : int
            If split_val is True, max_epochs_s1 and max_epochs_s2 correspond to 
            the maximum number of epochs for the two stages of training in 
            `train_with_val`, respectively. Otherwise, only max_epochs_s1 takes 
            effect, representing the maximum number of epochs in `train_without_val`. 
            Default is 1500 and 600 respectively.
        no_increase_epochs : int
            Maximum number of consecutive epochs when the accuracy or loss of 
            the training set (or val set if split_val is True) has no relative 
            improvement. If set to the same as max_epochs_s1, early stopping 
            will not be performed.
        var_check : str
            The best value (train_inacc/train_loss if split_val is False; val_
            inacc/val_loss if split_val is True) to check while determining the 
            best model which will be used to evaluate its performance on the 
            test set. Default is 'val_inacc' (if split_val is True) or 'train_
            loss' (if split_val is False).
        split_val : bool
            If True, the training set will be split into training set and 
            validation set, and early stopping will be performed based on the
            corresponding metrics. Otherwise, early stopping will be performed 
            based on the results of the training set. If False, ignore paramet-
            ers test_size, max_epochs_s2, second_stage, shuffle and seed.
        test_size : float
            The proportion of the validation set.
        second_stage : bool
            If True, two-stage training will be performed.
        load_best_state : bool
            If True, stage 2 will retrain based on the best state in stage 1 
            when second_stage is True or the best model parameters will be used
            for evaluation when second_stage is False.
        seed : int
            Seed of random for review.
        '''
        super().__init__(trainer, out_folder, verbose)

        self._repr = get_init_args(self, locals())
        self.max_epochs_s1 = max_epochs_s1
        self.no_increase_epochs = no_increase_epochs
        if var_check:
            self.var_check = var_check
        else:
            self.var_check = 'val_inacc' if split_val else 'train_loss'
        self.split_val = split_val
        self.test_size = test_size
        self.second_stage = second_stage
        self.load_best_state = load_best_state
        self.max_epochs_s2 = max_epochs_s2
        self.seed = seed

    def _run_sub(
        self,
        trainset : Union[tuple, list],
        testset : Union[tuple, list],
        cls_name : Union[tuple, list],
        sub : str,
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, dict]:
        '''Basic holdout cross validation function.
        '''
        # set subject's reults storage path
        sub_folder = os.path.join(self.data_folder, sub)
        os.makedirs(sub_folder)

        # run hold-out
        self.loger.info(f'\n# ------------ {sub} ------------ #')
        if self.split_val:
            trainX, valX, trainy, valy = split_train_test(
                *trainset, test_size=self.test_size, seed=self.seed
            )
            result = self.trainer.fit_with_val(
                trainset=(trainX, trainy),
                valset=(valX, valy),
                testset=testset,
                log_dir=sub_folder,
                cls_name=cls_name,
                max_epochs_s1=self.max_epochs_s1,
                max_epochs_s2=self.max_epochs_s2,
                no_increase_epochs=self.no_increase_epochs,
                second_stage=self.second_stage,
                load_best_state=self.load_best_state,
                var_check=self.var_check # type: ignore
            )
        else:
            result = self.trainer.fit_without_val(
                trainset=trainset,
                testset=testset,
                log_dir=sub_folder,
                cls_name=cls_name,
                max_epochs=self.max_epochs_s1,
                no_increase_epochs=self.no_increase_epochs,
                var_check=self.var_check, # type: ignore
                load_best_state=self.load_best_state
            )

        # save all results
        train_res, test_res = result['train'], result['test']
        train_acc, test_acc = train_res['acc'], test_res['acc']
        test_preds, test_target = test_res['preds'], test_res['target']
        save_cm_img(test_preds, test_target, cls_name, 
                    os.path.join(sub_folder, f'{sub}_CM.png')
        )
        self.loger.info(f'Acc : train={train_acc:.4f} | test={test_acc:.4f}')
        test_kappa = cohen_kappa(test_preds, test_target, task='multiclass',
                                 num_classes=len(cls_name))

        return test_acc, test_kappa, test_preds, test_target, result


class LOSO(Experiment):
    '''Leave-One-Subject-Out base class.

    In each fold of LOSO-CV, a single subject was used as the testing set, and
    the remaining N - 1 subjects were employed as the training set to obtain N 
    classification results. The training set was constructed using all data 
    sessions of all N - 1 training subjects for each dataset.
    '''
    def _merge_sub_dataset(self, exc_sub : int):
        '''Merge subject data except `sub`.
        '''
        merge_dataset = {}
        for sub, data in self.dataset.items():
            if sub != exc_sub:
                self.loger.debug(f'Merge sub_{sub} train and test set.')
                merge_dataset[sub] = merge_train_test(*data.values())

        self.loger.debug('Merge all subject dataset.')
        merge_dataset = merge_train_test(*merge_dataset.values())
        return merge_dataset

    def _process_sub_dataset(self, sub : int) -> tuple:
        '''Merge data sets according to LOSO experiment requirements.
        '''
        self.loger.info(f'[Leave Subject {sub} Out]')
        testset = merge_train_test(*self.dataset[sub].values())
        trainset = self._merge_sub_dataset(sub)
        trainset, testset = self._trans_dataset(trainset, testset)

        return trainset, testset


class LOSO_HO(LOSO, Holdout):
    '''Leave-One-Subject-Out Holdout cross validation experiment.
    '''
    pass


class LOSO_CV(LOSO, KFold):
    '''Leave-One-Subject-Out K-Fold cross validation experiment.
    '''
    def _process_sub_dataset(self, sub : int) -> tuple:
        if self.isolate_testset == False:
            raise  RuntimeError('LOSO_CV only support isolate testset.')
        return super()._process_sub_dataset(sub)


import torch
import torch.nn as nn
import torch.nn.functional as F
from einops.layers.torch import Rearrange


class TSConv(nn.Sequential):
    def __init__(self, nCh, F, C1, C2, D, P1, P2, Pc) -> None:
        super().__init__(
            nn.Conv2d(1, F, (1, C1), padding='same', bias=False),
            nn.BatchNorm2d(F),
            nn.Conv2d(F, F * D, (nCh, 1), groups=F, bias=False),
            nn.BatchNorm2d(F * D),
            nn.ELU(),
            nn.AvgPool2d((1, P1)),
            nn.Dropout(Pc),
            nn.Conv2d(F * D, F * D, (1, C2), padding='same', groups=F * D, bias=False),
            nn.BatchNorm2d(F * D),
            nn.ELU(),
            nn.AvgPool2d((1, P2)),
            nn.Dropout(Pc)
        )


class PositionalEncoding(nn.Module):
    def __init__(self, seq_len, d_model) -> None:
        super().__init__()
        self.seq_len = seq_len
        self.d_model = d_model
        self.pe = nn.Parameter(torch.zeros(1, seq_len, d_model))

    def forward(self, x):
        x += self.pe
        return x
        

class Transformer(nn.Module):
    def __init__(
        self,
        seq_len,
        d_model, 
        nhead, 
        ff_ratio, 
        Pt = 0.5, 
        num_layers = 4, 
    ) -> None:
        super().__init__()
        self.cls_embedding = nn.Parameter(torch.zeros(1, 1, d_model))
        self.pos_embedding = PositionalEncoding(seq_len + 1, d_model)

        dim_ff =  d_model * ff_ratio
        self.dropout = nn.Dropout(Pt)
        self.trans = nn.TransformerEncoder(nn.TransformerEncoderLayer(
            d_model, nhead, dim_ff, Pt, batch_first=True, norm_first=True
        ), num_layers, norm=nn.LayerNorm(d_model))

    def forward(self, x):
        b = x.shape[0]
        x = torch.cat((self.cls_embedding.expand(b, -1, -1), x), dim=1)
        x = self.pos_embedding(x)
        x = self.dropout(x)
        return self.trans(x)[:, 0]


class ClsHead(nn.Sequential):
    def __init__(self, linear_in, cls):
        super().__init__(
            nn.Flatten(),
            nn.Linear(linear_in, cls),
            nn.LogSoftmax(dim=1)
        )


class MSVTNet(nn.Module):
    def __init__(
        self,
        nCh = 22,
        nTime = 1000,
        cls = 4,
        F = [9, 9, 9, 9],
        C1 = [15, 31, 63, 125],
        C2 = 15,
        D = 2,
        P1 = 8,
        P2 = 7,
        Pc = 0.3,
        nhead = 8,
        ff_ratio = 1,
        Pt = 0.5,
        layers = 2,
        b_preds = True,
    ) -> None:
        super().__init__()
        self.nCh = nCh
        self.nTime = nTime
        self.b_preds = b_preds
        assert len(F) == len(C1), 'The length of F and C1 should be equal.'

        self.mstsconv = nn.ModuleList([
            nn.Sequential(
                TSConv(nCh, F[b], C1[b], C2, D, P1, P2, Pc),
                Rearrange('b d 1 t -> b t d')
            )
            for b in range(len(F))
        ])
        branch_linear_in = self._forward_flatten(cat=False)
        self.branch_head = nn.ModuleList([
            ClsHead(branch_linear_in[b].shape[1], cls)
            for b in range(len(F))
        ])

        seq_len, d_model = self._forward_mstsconv().shape[1:3] # type: ignore
        self.transformer = Transformer(seq_len, d_model, nhead, ff_ratio, Pt, layers)

        linear_in = self._forward_flatten().shape[1] # type: ignore
        self.last_head = ClsHead(linear_in, cls)

    def _forward_mstsconv(self, cat = True):
        x = torch.randn(1, 1, self.nCh, self.nTime)
        x = [tsconv(x) for tsconv in self.mstsconv]
        if cat:
            x = torch.cat(x, dim=2)
        return x

    def _forward_flatten(self, cat = True):
        x = self._forward_mstsconv(cat)
        if cat:
            x = self.transformer(x)
            x = x.flatten(start_dim=1, end_dim=-1)
        else:
            x = [_.flatten(start_dim=1, end_dim=-1) for _ in x]
        return x

    def forward(self, x):
        x = [tsconv(x) for tsconv in self.mstsconv]
        bx = [branch(x[idx]) for idx, branch in enumerate(self.branch_head)]
        x = torch.cat(x, dim=2)
        x = self.transformer(x)
        x = self.last_head(x)
        if self.b_preds:
            return x, bx
        else:
            return x


class JointCrossEntoryLoss(nn.Module):
    def __init__(self, lamd : float = 0.6) -> None:
        super().__init__()
        self.lamd = lamd

    def forward(self, out, label):
        end_out = out[0]
        branch_out = out[1]
        end_loss = F.nll_loss(end_out, label)
        branch_loss = [F.nll_loss(out, label).unsqueeze(0) for out in branch_out]
        branch_loss = torch.cat(branch_loss)
        loss = self.lamd * end_loss + (1 - self.lamd) * torch.sum(branch_loss)
        return loss


if __name__ == '__main__':
    from torchinfo import summary
    net = MSVTNet(nCh=22, nTime=1000).cuda()
    print(net)
    summary(net, (64, 1, 22, 1000), depth=4)


'''
References
----------
R. T. Schirrmeister et al., “Deep learning with convolutional neural networks 
for EEG decoding and visualization,” Human Brain Mapping, vol. 38, no. 11, pp.
5391-5420, 2017, doi: 10.1002/hbm.23730.
'''


import torch
import torch.nn as nn

from .utils import Conv2dWithNorm, LinearWithNorm


class Lambda(nn.Module):
    def __init__(self, func) -> None:
        super().__init__()
        self.func = func

    def forward(self, x):
        return self.func(x)


class ShallowConvNet(nn.Module):
    def __init__(self, nCh=22, nTime=1000, cls=4, F=40, C=14, P=35, S=7, 
                 dropout=0.5) -> None:
        super().__init__()
        self.nCh = nCh
        self.nTime = nTime

        self.conv = nn.Sequential(
            Conv2dWithNorm(1, F, (1, 14), max_norm=2, bias=False),
            Conv2dWithNorm(F, F, (nCh, 1), max_norm=2, bias=False, groups=F),
            nn.BatchNorm2d(F),
            Lambda(torch.square),
            nn.AvgPool2d((1, P), stride=(1, S)),
            Lambda(torch.log),
        )

        linear_in = self.forward_flatten().shape[1]
        self.head = nn.Sequential(
            nn.Flatten(),
            nn.Dropout(dropout),
            LinearWithNorm(linear_in, cls, max_norm=0.5),
            nn.LogSoftmax(dim=1)
        )

    def forward_flatten(self):
        x = torch.rand(1, 1, self.nCh, self.nTime)
        x = self.conv(x)
        x = torch.flatten(x, 1, -1)
        return x

    def forward(self, x):
        x = self.conv(x)
        x = self.head(x)
        return x


if __name__ == '__main__':
    from torchinfo import summary
    net = ShallowConvNet().cuda()
    summary(net, (32, 1, 22, 1000))
    


'''
References
----------
R. T. Schirrmeister et al., “Deep learning with convolutional neural networks 
for EEG decoding and visualization,” Human Brain Mapping, vol. 38, no. 11, pp.
5391-5420, 2017, doi: 10.1002/hbm.23730.
'''


import torch
import torch.nn as nn

from .utils import Conv2dWithNorm, LinearWithNorm


class DeepConvNet(nn.Module):
    def __init__(self, nCh=22, nTime=1000, cls=4, dropout=0.25) -> None:
        super().__init__()
        self.nCh = nCh
        self.nTime = nTime
        kernel_size = [1, 10]
        filter_layer = [25, 50, 100, 200]

        first_layer = nn.Sequential(
            Conv2dWithNorm(1, 25, kernel_size, max_norm=2),
            Conv2dWithNorm(25, 25, (nCh, 1), bias=False, max_norm=2),
            nn.BatchNorm2d(25),
            nn.ELU(),
            nn.MaxPool2d((1, 3))
        )
        middle_layer = nn.Sequential(*[
            nn.Sequential(
                nn.Dropout(dropout),
                Conv2dWithNorm(in_f, out_f, kernel_size),
                nn.BatchNorm2d(out_f),
                nn.ELU(),
                nn.MaxPool2d((1, 3))
            ) for in_f, out_f in zip(filter_layer, filter_layer[1:])
        ])
        self.conv_layer = nn.Sequential(first_layer, middle_layer)

        linear_in = self._forward_flatten().shape[1]
        self.head = nn.Sequential(
            nn.Flatten(),
            LinearWithNorm(linear_in, cls, max_norm=0.5),
            nn.LogSoftmax(dim=1)
        )

    def _forward_flatten(self):
        x = torch.rand(1, 1, self.nCh, self.nTime)
        x = self.conv_layer(x)
        x = torch.flatten(x, start_dim=1, end_dim=-1)
        return x

    def forward(self, x):
        x = self.conv_layer(x)
        x = self.head(x)
        return x


if __name__ == '__main__':
    from torchinfo import summary
    net = DeepConvNet(22, 1000).cuda()
    summary(net, (64, 1, 22, 1000))


#!/usr/bin/env python
# coding: utf-8

"""
    Mainly apply corresponding mne library functions to implement preprocessing
    before extracting the data for Epochs data.

    @Author  : SheepTAO
    @Time    : 2023-10-05
"""


import abc
from typing import Optional, Union, List

from ..utils import dict_to_str, unpacked


class Preprocess(abc.ABC):
    @abc.abstractmethod
    def __call__(self, input : dict) -> dict:
        pass

    @abc.abstractmethod
    def __repr__(self) -> str:
        pass


class ComposePreprocess(Preprocess):
    def __init__(
        self, 
        *preprocess : Preprocess, 
    ) -> None:
        '''Compose serval preprocess together.

        The preprocess in `ComposePreprocess` are connected in a cascading way.

        Parameters
        ----------
        preprocess : sequential container of Preprocess
            Preprocess (sequential of `Preprocess` objects): sequential of pre-
            process to compose.
        '''
        self.preps : List[Preprocess] = []
        self.appends(*preprocess)

    def __call__(self, input: dict) -> dict:
        for t in self.preps:
            input = t(input)
        return input

    def __repr__(self) -> str:
        s = 'ComposePreprocess('
        if len(self.preps) == 0:
            return s + ')'
        else:
            for idx, pre in enumerate(self.preps):
                s += f'\n ({idx}): {pre}'
        return s + '\n)'

    def appends(self, *preprocess : Preprocess) -> None:
        '''Append preprocess to the last of composes.
        '''
        preps = unpacked(*preprocess)
        for prep in preps:
            if isinstance(prep, ComposePreprocess):
                self.preps.extend(prep.get_data())
            else:
                self.preps.append(prep)

    def insert(self, index : int, preprocess : Preprocess) -> None:
        '''Insert a preprocess at index.
        '''
        self.preps.insert(index, preprocess)

    def get_data(self) -> List[Preprocess]:
        '''Return list of Preprocess.
        '''
        return self.preps


class Filter(Preprocess):
    def __init__(
        self, 
        lfreq : Optional[float] = None,
        hfreq : Optional[float] = None,
        **mne_filter_kwargs,
    ) -> None:
        '''Applies filter to the signals in epochs. The epoch will be modified 
        in-place.

        Parameters
        ----------
        lfreq : float, optional
            For FIR filters, the lower pass-band edge; for IIR filters, the 
            lower cutoff frequency. If None the data are only low-passed.
        hfreq : float, optional
            For FIR filters, the upper pass-band edge; for IIR filters, the up-
            per cutoff frequency. If None the data are only high-passed.
        mne_filter_kwargs : dict
            Keyword arguments for filtering supported by `mne.io.Epochs.filter`
            Please refer to mne for a detailed explanation.
        '''
        self.lfreq = lfreq
        self.hfreq = hfreq
        self.mne_filter_kwargs = mne_filter_kwargs

    def __call__(self, input: dict) -> dict:
        for sub in input.values():
            if isinstance(sub, dict):
                for epoch in sub.values():
                    epoch.filter(
                        self.lfreq, self.hfreq, **self.mne_filter_kwargs
                    )
            else:
                sub.filter(self.lfreq, self.hfreq, **self.mne_filter_kwargs)
        return input

    def __repr__(self) -> str:
        s = f'Filter(lFreq={self.lfreq}, hFreq={self.hfreq}'
        if self.mne_filter_kwargs:
            s += f', {dict_to_str(self.mne_filter_kwargs)}'
        return s + ')'


class Resample(Preprocess):
    def __init__(
        self,
        sfreq : float,
        **mne_resample_kwargs 
    ) -> None:
        '''Resample data in epochs. The epoch will be modified in-place.

        Parameters
        ----------
        sfreq : float
            New sample rate to use.
        mne_resample_kwargs : dict
            Keyword arguments for resample supported by `mne.io.Epochs.resample`
            Please refer to mne for a detailed explanation.
        '''
        self.sfreq = sfreq
        self.mne_resample_kwargs = mne_resample_kwargs

    def __call__(self, input: dict) -> dict:
        for sub in input.values():
            if isinstance(sub, dict):
                for epoch in sub.values():
                    epoch.resample(self.sfreq, **self.mne_resample_kwargs)
            else:
                sub.resample(self.sfreq, **self.mne_resample_kwargs)
        return input

    def __repr__(self) -> str:
        s = f'Resample(sfreq={self.sfreq}'
        if self.mne_resample_kwargs:
            s += f', {dict_to_str(self.mne_resample_kwargs)}'
        return s + ')'



#!/usr/bin/env python
# coding: utf-8

"""
    This module is used to perform common preprocessing on eeg data (ndarray). 
    All transforms is in-place.

    TODO: Data augmentation.
    
    @author: SheepTAO
    @data: 2023-5-15
"""


import abc
import numpy as np
import pandas as pd
from typing import Optional, Callable, List

from ..utils import loger, verbose, DPEEG_SEED, unpacked, get_init_args
from ..tools.logger import _Level
import dpeeg.data.functions as F
from .functions import (
    split_train_test,
    to_tensor,
    slide_win,
    save,
    cheby2_filter,
)


class Transforms(abc.ABC):
    def __init__(self) -> None:
        self._repr = None
    
    @abc.abstractmethod
    def __call__(self, input : dict, verbose : _Level = None) -> dict:
        pass

    def __repr__(self) -> str:
        if self._repr:
            return self._repr
        else:
            class_name = self.__class__.__name__
            return f'{class_name} not implement attribute `self._repr`.'


class ComposeTransforms(Transforms):
    def __init__(self, *transforms : Transforms) -> None:
        '''Composes several transforms together. 

        The transforms in `ComposeTransforms` are connected in a cascading way.

        Parameters
        ----------
        transforms : sequential container of `Transforms`
            Sequential of transforms to compose. 

        Examples
        --------
        If you have multiple transforms that are processed sequentiallt, you 
        can do like:
        >>> from dpeeg.data import transforms
        >>> trans = transforms.ComposeTransforms(
        ...     transforms.Normalization(),
        ...     transforms.Unsqueeze(),
        ... )
        >>> trans
        ComposeTransforms(
            (0): Normalization(mode=z-score)
            (1): Unsqueeze(dim=1)
        )
        '''
        super().__init__()
        self.trans : List[Transforms] = []
        self.appends(*transforms)

    @verbose
    def __call__(self, input : dict, verbose : _Level = None):
        for tran in self.trans:
            input = tran(input, verbose=verbose)
        return input

    def __repr__(self) -> str:
        s = 'ComposeTransforms('
        if len(self.trans) == 0:
            return s + ')'
        else:
            for idx, tran in enumerate(self.trans):
                s += f'\n ({idx}): {tran}'
        return s + '\n)'

    def appends(self, *transforms : Transforms) -> None:
        '''Append transforms to the last of composes.
        '''
        trans = unpacked(*transforms)
        for tran in trans:
            if isinstance(tran, ComposeTransforms):
                self.trans.extend(tran.get_data())
            else:
                self.trans.append(tran)
        
    def insert(self, index : int, transform : Transforms) -> None:
        '''Insert a transform at index.
        '''
        self.trans.insert(index, transform)

    def get_data(self) -> List[Transforms]:
        '''Return list of Transforms.
        '''
        return self.trans


class SplitTrainTest(Transforms):
    def __init__(
        self, 
        test_size : float = .25, 
        seed : int = DPEEG_SEED, 
        sample : Optional[List[int]] = None,
    ) -> None:
        '''Split the dataset into training and testing sets.

        Parameters
        ----------
        test_size : float
            The proportion of the test set. If index not None, test_size will 
            be ignored. Default use stratified fashion and the last arr serves 
            as the class labels.
        seed : int
            Random seed when splitting.
        sample : list of int, optional
            A list of integers, the entries indicate which data were selected 
            as the test set. If None, test_size will be used.
        '''
        super().__init__()
        self._repr = get_init_args(self, locals(), format='rp')
        self.test_size = test_size
        self.seed = seed
        self.sample = sample

    @verbose
    def __call__(self, input : dict, verbose : _Level = None) -> dict:
        loger.info(f'[{self} starting] ...')
        for sub, data in input.items():
            trainX, testX, trainy, testy = split_train_test(
                data[0], data[1], test_size=self.test_size, seed=self.seed,
                sample=self.sample, verbose=verbose
            )
            input[sub] = {}
            input[sub]['train'] = [trainX, trainy]
            input[sub]['test'] = [testX, testy]
        return input


class ToTensor(Transforms):
    '''Convert the numpy data in the dataset into Tensor format.
    '''
    @verbose
    def __call__(self, input : dict, verbose : _Level = None) -> dict:
        loger.info(f'[{self} starting] ...')
        for sub in input.values():
            for mode in ['train', 'test']:
                sub[mode] = list(to_tensor(*sub[mode], verbose=verbose))
        return input

    def __repr__(self) -> str:
        return 'ToTensor'


class Normalization(Transforms):
    def __init__(
        self, 
        mode : str = 'z-score', 
        factor_new : float = 1e-3,
        verbose : _Level = None
    ) -> None:
        '''Normalize data in the given way in the given dimension.

        Parameters
        ----------
        mode : str
            within subject:
            - `z-score`,
                :math: $X_{i}=\\frac{X_{i}-mean(X_{i})}{std(X_{i})}$ where mean
                represents the channel-wise average value and std represents the 
                channel-wise standard deviation value, calculated with the training
                data and used directly for the test data.
            - `ems` (exponential moving standardize),
                Compute the exponental moving mean :math: $m_t$ at time $t$ as
                :math: $m_t=\\mathrm{factornew}\\cdot mean(x_t)+(1-\\mathrm{factornew})\\cdot m_{t-1}$.
                Then compute exponential moving variance $v_t$ at time t as
                :math: $v_t=\\mathrm{factornew}\\cdot (m_t-x_t)^2+(1-\\mathrm{factornew})\\cdot v_{t-1}$.
                Finally, standardize the data point :math: $x_t$ at time $t$ as:
                :math: $x'_t=(x_t-m_t)/max(\\sqrt{->v_t}, eps)$.
            cross subject:
            - `ea` (euclidean-space alignment),
                An unsupervised standardization of data transfer across subjects.
                > H. He and D. Wu, 'Transfer Learning for Brain-Computer Interfaces:
                A Euclidean Space Data Alignment Approach', Ieee T Bio-med Eng, vol. 
                67, no. 2, pp. 399-410, Feb. 2020, doi: 10.1109/TBME.2019.2913914.
            
        factor_new : float
            Smoothing factor of exponential moving standardize.

        Notes
        -----
        heavy development.
        '''
        super().__init__()
        self._repr = get_init_args(self, locals(), format='rp')
        self.mode_list = ['z-score', 'ems', 'ea']
        self.mode = mode
        self.factor_new = factor_new

    @verbose
    def __call__(self, input : dict, verbose : _Level = None) -> dict:
        if self.mode not in self.mode_list:
            raise ValueError('Only the following normalization methods are '+
                             f'supported: {self.mode_list}')

        loger.info(f'[{self} starting] ...')
        R = np.empty(0)
        if self.mode == 'ea':
            dataList = []
            for sub in input.values():
                for mode in ['train', 'test']:
                    dataList.append(sub[mode][0] / np.max(np.abs(sub[mode][0])))
            data = np.concatenate(dataList)

            Rbar = np.zeros((data.shape[1], data.shape[1]))
            for i in range(len(data)):
                Rbar += np.dot(data[i], data[i].T)
            RbarMean = Rbar / len(data)
            R = np.linalg.inv(np.sqrt(RbarMean))

        for sub in input.values():

            if self.mode == 'z-score':
                m, s = np.mean(sub['train'][0]), np.std(sub['train'][0])
                for mode in ['train', 'test']:
                    sub[mode][0] = (sub[mode][0] - m) / s

            # BUG
            elif self.mode == 'ems':
                df = pd.DataFrame(sub['train'][0].T)
                meaned = df.ewm(alpha=self.factor_new).mean()
                for mode in ['train', 'test']:
                    df = pd.DataFrame(sub[mode][0].T)
                    demeaned = df - meaned
                    squared = demeaned * demeaned
                    squareEwmed = squared.ewm(alpha=self.factor_new).mean()
                    standardized = demeaned / np.sqrt(np.array(squareEwmed))
                    standardized = np.array(standardized)
                    sub[mode][0] = standardized.T

            # NOT TEST
            elif self.mode == 'ea':
                for mode in ['train', 'test']:
                    sub[mode][0] /= np.max(np.abs(sub[mode][0]))
                    sub[mode][0] = np.dot(sub[mode][0], R)
        return input


class SlideWin(Transforms):
    def __init__(self, win : int, overlap : int = 0) -> None:
        '''Apply a sliding window to the dataset.

        This transform is only splits the time series (dim = -1) through the 
        sliding window operation on the original dataset. If the time axis is 
        not divisible by the sliding window, the last remaining time data will 
        be discarded.

        Parameters
        ----------
        win : int
            The size of the sliding window.
        overlap : int
            The amount of overlap between adjacent sliding windows.
        '''
        super().__init__()
        self._repr = get_init_args(self, locals(), format='rp')
        self.win = win
        self.overlap = overlap

    @verbose
    def __call__(self, input : dict, verbose : _Level = None) -> dict:
        loger.info(f'[{self} starting] ...')
        for sub in input.values():
            for mode in ['train', 'test']:
                sub[mode][0], sub[mode][1] = slide_win(
                    sub[mode][0], self.win, self.overlap,
                    sub[mode][1], verbose=verbose
                )
        return input


class Unsqueeze(Transforms):
    def __init__(self, dim : int = 1) -> None:
        '''Insert a dimension on the data.

        This transform is usually used to insert a empty dimension on EEG data.

        Parameters
        ----------
        dim : int
            Position in the expanded dim where the new dim is placed.
        '''
        super().__init__()
        self._repr = get_init_args(self, locals(), format='rp')
        self.dim = dim

    @verbose
    def __call__(self, input : dict, verbose : _Level = None) -> dict:
        loger.info(f'[{self} starting] ...')
        for sub in input.values():
            for mode in ['train', 'test']:
                sub[mode][0] = np.expand_dims(sub[mode][0], self.dim)
        return input


class Augmentation(Transforms):
    def __init__(
        self,
        method : str,
        only_train : bool = True,
        **kwargs
    ) -> None:
        '''Training set data augmentation.

        This transform is mainly used for data augmentation of the data set.

        Parameters
        ----------
        method : str
            Specified data augmentation method.
        only_train : bool
            If True, data augmentation is performed only on the training set.
        kwargs : dict
            Parameters of the corresponding augmentation method.

        See Also
        --------
        dpeeg.data.functions : 
            Get detailed data augmentation method parameters.
        '''
        super().__init__()
        self._repr = get_init_args(self, locals(), format='rp')
        self.aug = getattr(F, method)
        self.only_train = only_train
        self.kwargs = kwargs

    @verbose
    def __call__(self, input : dict, verbose : _Level = None) -> dict:
        loger.info(f'[{self} starting] ...')
        for sub in input.values():
            for mode in ['train', 'test']:
                if mode == 'test' and self.only_train:
                    continue
                sub[mode] = list(self.aug(*sub[mode], **self.kwargs))
        return input


class FilterBank(Transforms):
    def __init__(
        self,
        freq : float,
        filter_bank : list,
        transition_bandwidth : float = 2.,
        gstop : float = 30,
        gpass : float = 3,
    ) -> None:
        '''Filter Bank.

        EEG data will be filtered according to different filtering frequencies 
        and finally concatenated together. eg.`(Batch, ...) -> (Batch, F, ...)`
        if the number of filter banks exceeds 1, `(Batch, ...) -> (Batch, ...)`
        if the filter has only one.

        Parameters
        ----------
        freq : float
            Data sampling frequency.
        filter_bank : multiple 2 float of list
            The low-pass and high-pass cutoff frequencies for each filter set.
        transition_bandwidth : float
            The bandwidth (in hertz) of the transition region of the frequency 
            response from the passband to the stopband.
        gstop : float
            The minimum attenuation in the stopband (dB).
        gpass : float
            The maximum loss in the passband (dB).
        '''
        super().__init__()
        self._repr = get_init_args(self, locals(), format='rp')
        self.freq = freq
        self.filter_bank = self._check_filter_bank(filter_bank)
        self.transition_bandwidth = transition_bandwidth
        self.gpass = gpass
        self.gstop = gstop

    def _check_filter_bank(self, fb):
        if not isinstance(fb, list):
            raise TypeError(f'Filter_bank must be a list, not {type(fb)}.')
        for f in fb:
            if len(f) != 2:
                raise ValueError(
                    'The filter should be of two variables low pass and high '
                    'pass cutoff frequency.'
                )
        return fb

    @verbose
    def __call__(self, input: dict, verbose: _Level = None) -> dict:
        loger.info(f'[{self} starting] ...')
        for sub, sub_data in input.items():
            for mode in ['train', 'test']:

                bank_len = len(self.filter_bank)
                total_trials = F.total_trials(input, sub, mode) # type: ignore
                data = np.empty(
                    (total_trials, bank_len, *sub_data[mode][0].shape[1:])
                )

                for i, cutoff in enumerate(self.filter_bank):
                    filter_data = cheby2_filter(
                        data=sub_data[mode][0],
                        freq=self.freq,
                        l_freq=cutoff[0],
                        h_freq=cutoff[1],
                        transition_bandwidth=self.transition_bandwidth,
                        gpass=self.gpass,
                        gstop=self.gstop,
                        verbose=verbose,
                    )
                    data[:, i] = filter_data

                if bank_len == 1:
                    data = np.squeeze(data, 1)
                sub_data[mode][0] = data

        return input


class ApplyFunc(Transforms):
    def __init__(self, func : Callable[[np.ndarray], np.ndarray]) -> None:
        '''Apply a function on data.

        This transform can be used to change the data shape and so on.

        Examples
        --------
        If you want to pass a function with parameters, such as you want to use 
        `np.expand_dims()` with `axis` parameter, you can do as follows:
        >>> from functools import partial
        >>> transforms.ApplyFunc(partial(np.expand_dims, axis=1))

        or

        >>> transforms.ApplyFunc(lambda x: np.expand_dims(x, 1))
        '''
        super().__init__()
        self._repr = get_init_args(self, locals(), format='rp')
        self.func = func

    @verbose
    def __call__(self, input : dict, verbose : _Level = None) -> dict:
        loger.info(f'[{self} starting] ...')
        for sub in input.values():
            for mode in ['train', 'test']:
                sub[mode][0] = self.func(sub[mode][0])
        return input


class Save(Transforms):
    def __init__(self, folder : str) -> None:
        '''Save the transformed data.

        Save transformed dataset to a binary file in NumPy `.npy` format.

        Parameters
        ----------
        folder : str
            Folder name to save transformed data.
        '''
        super().__init__()
        self._repr = get_init_args(self, locals(), format='rp')
        self.folder = folder

    @verbose
    def __call__(self, input : dict, verbose : _Level = None) -> None:
        loger.info(f'[{self} starting] ...')
        save(self.folder, input, verbose=verbose)



import torch
import torch.nn.functional as F
import torch.nn as nn
from torch import Tensor
from einops import rearrange
from einops.layers.torch import Rearrange, Reduce


class PatchEmbedding(nn.Module):
    def __init__(self, emb_size=40):
        # self.patch_size = patch_size
        super().__init__()

        self.shallownet = nn.Sequential(
            nn.Conv2d(1, 40, (1, 25), (1, 1)),
            nn.Conv2d(40, 40, (22, 1), (1, 1)),
            nn.BatchNorm2d(40),
            nn.ELU(),
            nn.AvgPool2d((1, 75), (1, 15)),  # pooling acts as slicing to obtain 'patch' along the time dimension as in ViT
            nn.Dropout(0.5),
        )

        self.projection = nn.Sequential(
            nn.Conv2d(40, emb_size, (1, 1), stride=(1, 1)),  # transpose, conv could enhance fiting ability slightly
            Rearrange('b e (h) (w) -> b (h w) e'),
        )


    def forward(self, x: Tensor) -> Tensor:
        b, _, _, _ = x.shape
        x = self.shallownet(x)
        x = self.projection(x)
        return x


class MultiHeadAttention(nn.Module):
    def __init__(self, emb_size, num_heads, dropout):
        super().__init__()
        self.emb_size = emb_size
        self.num_heads = num_heads
        self.keys = nn.Linear(emb_size, emb_size)
        self.queries = nn.Linear(emb_size, emb_size)
        self.values = nn.Linear(emb_size, emb_size)
        self.att_drop = nn.Dropout(dropout)
        self.projection = nn.Linear(emb_size, emb_size)

    def forward(self, x: Tensor) -> Tensor:
        queries = rearrange(self.queries(x), "b n (h d) -> b h n d", h=self.num_heads)
        keys = rearrange(self.keys(x), "b n (h d) -> b h n d", h=self.num_heads)
        values = rearrange(self.values(x), "b n (h d) -> b h n d", h=self.num_heads)
        energy = torch.einsum('bhqd, bhkd -> bhqk', queries, keys)

        scaling = self.emb_size ** (1 / 2)
        att = F.softmax(energy / scaling, dim=-1)
        att = self.att_drop(att)
        out = torch.einsum('bhal, bhlv -> bhav ', att, values)
        out = rearrange(out, "b h n d -> b n (h d)")
        out = self.projection(out)
        return out


class ResidualAdd(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, **kwargs):
        res = x
        x = self.fn(x, **kwargs)
        x += res
        return x


class FeedForwardBlock(nn.Sequential):
    def __init__(self, emb_size, expansion, drop_p):
        super().__init__(
            nn.Linear(emb_size, expansion * emb_size),
            nn.GELU(),
            nn.Dropout(drop_p),
            nn.Linear(expansion * emb_size, emb_size),
        )


class TransformerEncoderBlock(nn.Sequential):
    def __init__(self,
                 emb_size,
                 num_heads=10,
                 drop_p=0.5,
                 forward_expansion=4,
                 forward_drop_p=0.5):
        super().__init__(
            ResidualAdd(nn.Sequential(
                nn.LayerNorm(emb_size),
                MultiHeadAttention(emb_size, num_heads, drop_p),
                nn.Dropout(drop_p)
            )),
            ResidualAdd(nn.Sequential(
                nn.LayerNorm(emb_size),
                FeedForwardBlock(
                    emb_size, expansion=forward_expansion, drop_p=forward_drop_p),
                nn.Dropout(drop_p)
            )
            ))


class TransformerEncoder(nn.Sequential):
    def __init__(self, depth, emb_size):
        super().__init__(*[TransformerEncoderBlock(emb_size) for _ in range(depth)])


class ClassificationHead(nn.Sequential):
    def __init__(self, emb_size, n_classes):
        super().__init__()
        
        # global average pooling
        self.clshead = nn.Sequential(
            Reduce('b n e -> b e', reduction='mean'),
            nn.LayerNorm(emb_size),
            nn.Linear(emb_size, n_classes)
        )
        self.fc = nn.Sequential(
            nn.Linear(2440, 256),
            nn.ELU(),
            nn.Dropout(0.5),
            nn.Linear(256, 32),
            nn.ELU(),
            nn.Dropout(0.3),
            nn.Linear(32, 4),
            nn.LogSoftmax(dim=1)
        )

    def forward(self, x):
        x = x.contiguous().view(x.size(0), -1)
        out = self.fc(x)
        return out


class EEGConformer(nn.Sequential):
    def __init__(self, emb_size=40, depth=6, n_classes=4, **kwargs):
        super().__init__(

            PatchEmbedding(emb_size),
            TransformerEncoder(depth, emb_size),
            ClassificationHead(emb_size, n_classes)
        )


if __name__ == '__main__':
    from torchinfo import summary
    net = EEGConformer()
    summary(net, (1, 1, 22, 1000), device='cpu')


#!/usr/bin/env python
# coding: utf-8

"""
    The train model for any deep learning analysis.
    This class should provide following functionalities for any deep learning.

    module:
        1. train() -> Train the model
        2. predict() -> Evaluate the train, validation, and test performance
        3. Create train and validation graphs
        4. Run over CPU / GPU (if available)
    This class needs following things to run:
        1. net -> The architecture of the network. It should inherit Module
             and should define the forward method.
        2. trainset, testset and valset -> these should be tuple, list or
             DataLoder the first parameters is `data` and second is `label`.
        3. optimizer -> the optimizer of type torch.optim.
        4. lr_scheduler -> the scheduler of type torch.optim.lr_scheduler.
        5. nGPU -> will run on GPU only if it's available, use CPU when the GPU
                is not available. 

    NOTE: The train model only support single-card training.

    @Author  : SheepTAO
    @Time    : 2023-07-24
"""


import os, torch
import torch.nn as nn
from torch import optim
from torch import Tensor
from torch.nn import Module
from copy import deepcopy
from torchinfo import summary
from typing import Optional, Tuple, Union, Dict, Literal
from torch.utils.tensorboard.writer import SummaryWriter
from torch.utils.data import TensorDataset, DataLoader, ConcatDataset
from torchmetrics.functional.classification.accuracy import multiclass_accuracy
from torchmetrics.aggregation import MeanMetric, CatMetric

from ..tools import Logger, Timer
from ..utils import DPEEG_SEED
from ..data.functions import to_tensor
from .stopcriteria import ComposeStopCriteria

from torch.backends import cudnn
cudnn.benchmark = False
cudnn.deterministic = True


# base train classifier 
class TrainClassifier:
    '''Apex Trainer for any deep learning.
    '''
    def __init__(
        self,
        net : Module,
        nGPU : int = 0,
        seed : int = DPEEG_SEED,
        loss_fn : Union[str, Module] = 'NLLLoss',
        loss_fn_args : Optional[dict] = None,
        optimizer : Union[str, Module] = 'Adam',
        optimizer_args : Optional[dict] = None,
        lr : float = 1e-3,
        lr_sch : Optional[Union[str, Module]] = None,
        lr_sch_args : Optional[dict] = None,
        grad_acc : int = 1,
        batch_size : int = 32,
        keep_data_gpu : bool = True,
        data_size : Optional[Union[tuple, list]] = None,
        depth : int = 3,
        verbose : Union[int, str] = 'INFO',
    ) -> None:
        '''Initialize the basic attribute of the train model.

        Generate a trainer to test the performance of the same network on 
        different datasets.

        Parameters
        ----------
        net : Module
            Inherit Module and should define the forward method. The first 
            parameter returned by model forward propagation is the prediction.
        nGPU : int
            Select the gpu id to train.
        seed : int
            Select random seed for review.
        loss_fn : str, Module
            Name of the loss function from torch.nn which will be used for
            training. If Module, means using a custom loss function.
        loss_fn_args : dict, optional
            Additional arguments to be passed to the loss function.
        optimizer : str, Module
            Name of the optimization function from torch.optim which will be
            used for training. If Module, means using a custom optimizer.
        optimizer_args : dict, optional
            Additional arguments to be passed to the optimization function.
        lr : float
            Learning rate.
        lr_sch : str, Module, optional
            Name of the lr_scheduler from torch.optim.lr_scheduler which will 
            be used for training. If Module, means using a custom lr_scheduler.
        lr_sch_args : dict, optional
            Additional arguments to be passed to the lr_scheduler function.
        grad_acc : int
            Aradient accumulation.
        batch_size : int
            Mini-batch size.
        keep_data_gpu : bool
            Keep the dataset on the GPU to avoid the time consumption of data 
            migration. Please adjust according to the personal GPU memory.
        data_size : tuple, list, optional
            Output the structure of the network model according to the input
            dimension if the `data_size` is given.
        depth : int
            Depth of nested layers to display.
        verbose : int, str
            The log level of console. Default is INFO. Mainly used for debug.
        '''
        self.net = net

        self.loger = Logger('dpeeg_train', clevel=verbose)
        self.timer = Timer()

        # init trainer
        self.device = self.get_device(nGPU)
        self.net.to(self.device)
        self.set_seed(seed)

        # summarize network structure
        self.net_arch = str(net) + '\n'
        self.net_arch += str(summary(net, data_size, depth=depth))
        self.loger.info(self.net_arch)

        # the type of optimizer, etc. selected
        self.loss_fn_type = loss_fn
        self.optimizer_type = optimizer
        self.lr_sch_type = lr_sch

        # save additional parameters
        self.loss_fn_args = loss_fn_args if loss_fn_args else {}
        self.optimizer_args = optimizer_args if optimizer_args else {}
        self.lr_sch_args = lr_sch_args if lr_sch_args else {}
        # --- others
        self.lr = lr
        self.seed = seed
        self.grad_acc = grad_acc
        self.batch_size = batch_size
        self.keep_data_gpu = keep_data_gpu
        self.verbose = verbose

        # set experimental details
        self.train_details = {
            'train_param': {
                'seed': seed, 
                'loss_fn': str(loss_fn), 
                'loss_fn_args': loss_fn_args, 
                'optimizer': str(optimizer), 
                'optimizer_args': optimizer_args, 
                'lr': lr,
                'lr_sch': str(lr_sch), 
                'lr_sch_args': lr_sch_args, 
                'batch_size': batch_size, 
                'grad_acc': grad_acc
            }, 
            'orig_net_param': deepcopy(self.net.state_dict()),
        }

    def _run_one_epoch(
        self,
        train_loader : DataLoader
    ) -> None:
        '''Run one epoch to train net.

        Parameters
        ----------
        train_loader : DataLoader
            DataLoader used for training.
        '''
        # set the network in training mode
        self.net.train()

        # iterater over all the data
        with torch.enable_grad():
            for idx, (data, label) in enumerate(train_loader):
                data, label = data.to(self.device), label.to(self.device)
                out = self.net(data)
                loss = self.loss_fn(out, label)
                loss.backward()
                # gradient accumulation
                if((idx + 1) % self.grad_acc == 0):
                    # 1 - update parameters
                    self.optimizer.step()
                    # 2 - zero the parameter gradients
                    self.optimizer.zero_grad()
            # update lr
            # Note: Learning rate scheduling should be applied after optimizer’s update
            if self.lr_sch:
                self.lr_sch.step()

    def _predict(
        self,
        data_loader : DataLoader,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        '''Predict the class of the input data.

        Parameters
        ----------
        data_loader : DataLoader
            Dataset used for prediction.

        Returns
        -------
        preds : Tensor
            Predicted labels, as returned by a classifier.
        target : Tensor
            Ground truth (correct) labels.
        loss : Tensor
            Average loss.
        '''
        # set the network in the eval mode
        self.net.eval()

        loss_sum = MeanMetric()
        preds, target = CatMetric(), CatMetric()

        # iterate over all the data
        with torch.no_grad():
            for data, label in data_loader:
                data, label = data.to(self.device), label.to(self.device)
                out = self.net(data)
                loss = self.loss_fn(out, label)
                loss_sum.update(loss.item(), data.size(0))
                # convert the output of soft-max to class label
                # save preds and actual label
                out = out[0] if isinstance(out, tuple) else out
                preds.update(torch.argmax(out, dim=1).detach().cpu())
                target.update(label.cpu())
        return preds.compute(), target.compute(), loss_sum.compute()

    def _move_datasets(self, *datasets) -> list:
        '''Move datasets to specified device.

        Parameters
        ----------
        datasets : sequence of data (N, ...) and labels (N,)
            Sequence consisting of each piece of data and label. Can be ndarray
            and tensor.
        '''
        if len(datasets) == 0:
            raise ValueError('At least one dataset required as input.')

        ret_datasets = []
        for i, dataset in enumerate(datasets):
            if not (isinstance(dataset, (list, tuple)) and len(dataset) == 2):
                raise ValueError(f'Missing data or label in {i}th group.')

            data, label = to_tensor(dataset[0], dataset[1])
            data, label = data.to(self.device), label.to(self.device)
            ret_datasets.append((data, label))

        return ret_datasets

    def _data_loader(self, *datasets) -> DataLoader:
        '''Wrap multiple sets of data and labels and return DataLoader.

        Parameters
        ----------
        datasets : sequence of data (N, ...) and labels (N,)
            Sequence consisting of each piece of data and label. Can be ndarray
            and tensor.
        '''
        if len(datasets) == 0:
            raise ValueError('At least one dataset required as input.')

        # dataset wrapping tensors
        td = []
        for i, dataset in enumerate(datasets):
            if not (isinstance(dataset, (list, tuple)) and len(dataset) == 2):
                raise ValueError(f'Missing data or label in {i}th group.')

            data, label = to_tensor(dataset[0], dataset[1])
            if self.keep_data_gpu:
                data, label = data.to(self.device), label.to(self.device)
            td.append(TensorDataset(*to_tensor(data, label)))
        td = ConcatDataset(td)

        if self.keep_data_gpu:
            data_loader = DataLoader(td, self.batch_size, True)
        else:
            data_loader = DataLoader(td, self.batch_size, True, pin_memory=True)
        return data_loader

    def reset_fitter(
        self,
        log_dir : str
    ) -> Tuple[str, SummaryWriter, Logger]:
        '''Reset the relevant parameters of the fitter.

        Reset the model's training parameters, learning rate schedule and opti-
        mizer etc. to their initialized state.

        Parameters
        ----------
        log_dir : str
            Directory location (support hierarchical folder structure) to save
            training log.

        Returns
        -------
        str, SummaryWriter, Logger:
            Return the absolute file path and a new SummaryWriter object and 
            logger manager for the fitter.
        '''
        # reset parameters of nn.Moudle and lr_scheduler
        self.net.load_state_dict(self.train_details['orig_net_param'])

        # create loss function
        self.loss_fn = getattr(nn, self.loss_fn_type)(**self.loss_fn_args) \
            if isinstance(self.loss_fn_type, str) else self.loss_fn_type

        # create optimizer
        if isinstance(self.optimizer_type, str):
            self.optimizer = getattr(optim, self.optimizer_type) \
                (self.net.parameters(), lr=self.lr, **self.optimizer_args)
        else:
            self.optimizer = self.optimizer_type(
                self.net.parameters(), lr=self.lr, **self.optimizer_args)

        # create lr_scheduler
        if isinstance(self.lr_sch_type, str):
            self.lr_sch = getattr(optim.lr_scheduler, self.lr_sch_type) \
                (self.optimizer, **self.lr_sch_args)
        elif isinstance(self.lr_sch_type, Module):
            self.lr_sch = self.lr_sch_type(self.optimizer, **self.lr_sch_args)
        else:
            self.lr_sch = None

        # create log writer
        log_dir = os.path.abspath(log_dir)
        writer = SummaryWriter(log_dir)
        loger = Logger(log_dir, path=os.path.join(log_dir, 'running.log'), 
                       flevel='INFO', clevel=self.verbose)
        return log_dir, writer, loger

    def get_acc(self, preds : Tensor, target : Tensor, ncls : int) -> Tensor:
        '''Easy for program to caculate the accuarcy.
        '''
        return multiclass_accuracy(preds, target, ncls, 'micro')

    def set_seed(self, seed : int = DPEEG_SEED) -> None:
        '''Sets the seed for generating random numbers for cpu and gpu.
        '''
        torch.manual_seed(seed)
        if self.device != torch.device('cpu'):
            torch.cuda.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
        self.loger.info(f'Set all random seed = {seed}')

    def get_device(self, nGPU : int = 0) -> torch.device:
        '''Get the device for training and testing.

        Parameters
        ----------
        nGPU : int
            GPU number to train on.
        '''
        if not torch.cuda.is_available():
            self.loger.info('GPU is not avaiable and the CPU will be used')
            dev = torch.device('cpu')
        else:
            if nGPU > torch.cuda.device_count() - 1:
                raise ValueError(f'GPU: {nGPU} does not exit.')
            dev = torch.device(f'cuda:{nGPU}')
            self.loger.info(f'Network will be trained in "cuda:{nGPU} ' +
                            f'({torch.cuda.get_device_name(dev)})"')
        return dev

    def fit_without_val(
        self,
        trainset : Union[tuple, list],
        testset : Union[tuple, list],
        log_dir : str,
        cls_name : Union[tuple, list],
        max_epochs : int = 1000,
        no_increase_epochs : int = 200,
        var_check : Literal['train_loss', 'train_inacc'] = 'train_loss',
        load_best_state : bool = True,
    ) -> Dict[str, Dict[str, Tensor]]:
        '''During different training model processes, early stopping mechanisms
        can be executed using the training set (validation set not required) to
        select the model.

        Parameters
        ----------
        trainset : tuple, list
            Dataset used for training. If type is tuple or list, dataset should
            be (data, labels).
        testset : tuple, list
            Dataset used to evaluate the model. If type is tuple or list, 
            dataset should be (data, labels).
        log_dir : str
            Directory location (support hierarchical folder structure) to save
            training log.
        cls_name : tuple, list
            The name of dataset labels.
        max_epochs : int
            Maximum number of epochs in training.
        no_increase_epochs : int
            Maximum number of consecutive epochs when the accuracy or loss of 
            the training set has no relative improvement.
        var_check : str
            The best value (train_inacc/train_loss) to check while determining 
            the best model which will be used to evaluate its performance on 
            the test set.
        load_best_state : bool
            If True, the best model parameters will be used for evaluation.

        Returns
        -------
        dict
            Return train and test results dict.
            {
                'train' : {'preds': Tensor, 'target': Tensor, 'acc': Tensor},
                'test'  : ...
            }
        '''
        log_dir, writer, loger = self.reset_fitter(log_dir)

        # check the best model
        best_var = float('inf')
        best_net_param = deepcopy(self.net.state_dict())

        check_list = ['train_inacc', 'train_loss']
        if var_check not in check_list:
            s = ', '.join(check_list)
            raise ValueError(f'Parameter `var_check` only supports: {s} '
                             f'when training without val, but got {var_check}')

        # initialize dataloader
        if self.keep_data_gpu:
            trainset, testset = self._move_datasets(trainset, testset)
        train_loader = self._data_loader(trainset)
        test_loader = self._data_loader(testset)
        ncls = len(cls_name)

        # start the training
        self.timer.start()
        loger.info(f'[Training...] - [{self.timer.ctime()}]')
        stopcri = ComposeStopCriteria({'Or': {
            'cri1' : {'MaxEpoch': {
                'max_epochs': max_epochs, 'var_name': 'epoch'
            }},
            'cri2': {'NoDecrease': {
                'num_epochs': no_increase_epochs, 'var_name': var_check
            }}
        }})
        self.train_details['fit'] = {
            'type': 'fit_without_val',
            'var_check': var_check,
            'stopcri': str(stopcri)
        }
        monitors = {'epoch': 0, 'train_loss': float('inf'), 'train_inacc': 1}

        while not stopcri(monitors):

            # train one epoch
            self._run_one_epoch(train_loader)
            monitors['epoch'] += 1

            # evaluate the training and validation accuracy
            train_preds, train_target, train_loss = self._predict(train_loader)
            train_acc = self.get_acc(train_preds, train_target, ncls)
            monitors['train_inacc'] = 1 - train_acc
            monitors['train_loss'] = train_loss

            # store loss and acc
            writer.add_scalars('train', {'loss': train_loss, 'acc': train_acc},
                               monitors['epoch'])
            loger.info(f'-->Epoch : {monitors["epoch"]}')
            loger.info(f'  \u21b3train Loss/Acc = {train_loss:.4f}/{train_acc:.4f}')

            # select best model
            if load_best_state and monitors[var_check] <= best_var:
                best_var = monitors[var_check]
                best_net_param = deepcopy(self.net.state_dict())

        writer.close()
        if not load_best_state:
            best_net_param = deepcopy(self.net.state_dict())

        # report the checkpoint time of end and compute cost time
        loger.info(f'[Train Finish] - [{self.timer.ctime()}]')
        h, m, s = self.timer.stop()
        loger.info(f'Cost time = {h}H:{m}M:{s:.2f}S')

        # load the best model and evaulate this model in testset
        self.net.load_state_dict(best_net_param)

        results = {}
        train_preds, train_target, train_loss = self._predict(train_loader)
        train_acc = self.get_acc(train_preds, train_target, ncls)
        results['train'] = {
            'preds': train_preds, 'target': train_target, 'acc': train_acc
        }
        test_preds, test_target, test_loss = self._predict(test_loader)
        test_acc = self.get_acc(test_preds, test_target, ncls)
        results['test'] = {
            'preds': test_preds, 'target': test_target, 'acc': test_acc
        }

        loger.info(f'Loss: train={train_loss:.4f} | test={test_loss:.4f}')
        loger.info(f'Acc:  train={train_acc:.4f} | test={test_acc:.4f}')

        # save the training results
        self.train_details['results'] = results
        self.train_details['best_net_param'] = best_net_param

        # store the training details
        train_details_path = os.path.join(log_dir, f'train_details.pt')
        torch.save(self.train_details, train_details_path)

        # store the best net model parameters
        best_checkpoiont_path = os.path.join(log_dir, f'best_checkpoint.pth')
        torch.save(best_net_param, best_checkpoiont_path)

        return results

    def fit_with_val(
        self,
        trainset : Union[tuple, list],
        valset : Union[tuple, list],
        testset : Union[tuple, list],
        log_dir : str,
        cls_name : Union[tuple, list],
        max_epochs_s1 : int = 1500,
        max_epochs_s2 : int = 600,
        no_increase_epochs : int = 200,
        second_stage : bool = True,
        load_best_state : bool = True,
        var_check : Literal['val_inacc', 'val_loss'] = 'val_inacc',
    ) -> Dict[str, Dict[str, Tensor]]:
        '''Two-stage training strategy was used. In the first stage, the model 
        was trained using only the training set with the early stopping criteria
        whereby the validation set accuracy and loss was monitored and training
        was stopped if there was no increase in the validation set accuracy (or 
        loss) for consecutive 200 epochs. After reaching the stopping criteria, 
        network parameters with the best validation set accuracy (or loss) were
        restored. In the second stage, the model was trained with the complete 
        training data (train + validation set). The second stage training was
        stopped when the validation set loss reduced below the stage 1 training 
        set loss.

        Parameters
        ----------
        trainset : tuple, list
            Dataset used for training. If type is tuple or list, dataset should
            be (data, labels).
        valset : tuple, list
            Dataset used for validation. If type is tuple or list, dataset 
            should be (data, labels).
        testset : tuple, list
            Dataset used to evaluate the model. If type is tuple or list, 
            dataset should be (data, labels).
        log_dir : str
            Directory location (support hierarchical folder structure) to save
            training log.
        cls_name : tuple, list
            The name of dataset labels.
        max_epochs_s1, max_epochs_s2 : int
            Maximum number of epochs in the x stage of training. Default is
            1500 and 600 respectively.
        no_increase_epochs : int
            Maximum number of consecutive epochs when the accuracy or loss of 
            the first-stage validation set has no relative improvement.
        second_stage : bool
            If True, two-stage training will be performed.
        load_best_state : bool
            If True, stage 2 will retrain based on the best state in stage 1.
        var_check : str
            The best value (val_inacc/val_loss) to check while determining the 
            best state which will be used for parameter initialization in the
            second stage of model training.

        Returns
        -------
        dict
            Return train, validation and test results dict.
            {
                'train' : {'preds': Tensor, 'target': Tensor, 'acc': Tensor},
                'test'  : ...,
                'val'   : ...
            }
        '''
        log_dir, writer, loger = self.reset_fitter(log_dir)

        # check the best model
        best_var = float('inf')
        best_net_param = deepcopy(self.net.state_dict())
        best_optim_param = deepcopy(self.optimizer.state_dict())

        check_list = ['val_inacc', 'val_loss']
        if var_check not in check_list:
            s = ', '.join(check_list)
            raise ValueError(f'Parameter `var_check` only supports: {s}'
                             f' when training with val, but got {var_check}')

        # initialize dataloader
        if self.keep_data_gpu:
            trainset, valset, testset = self._move_datasets(
                trainset, valset, testset
            )
        train_loader = self._data_loader(trainset)
        val_loader = self._data_loader(valset)
        test_loader = self._data_loader(testset)
        ncls = len(cls_name)

        # start the training
        self.timer.start()
        loger.info(f'[Training...] - [{self.timer.ctime()}]')

        stopcri = ComposeStopCriteria({'Or': {
            'cri1': {'MaxEpoch': {
                'max_epochs': max_epochs_s1, 'var_name': 'epoch'
            }},
            'cri2': {'NoDecrease': {
                'num_epochs': no_increase_epochs, 'var_name': var_check
            }}
        }})
        self.train_details['fit'] = {
            'type': 'fit_with_val',
            'var_check': var_check,
            'stopcri_1': str(stopcri)
        }
        monitors = {
            'epoch': 0, 
            'val_loss': float('inf'), 
            'val_inacc': 1,
            'global_epoch': 0,
            'best_epoch': -1,
            'best_train_loss': float('inf'), 
        }
        early_stop_reached, do_stop = False, False

        while not do_stop: 

            # train one epoch
            self._run_one_epoch(train_loader)
            monitors['epoch'] += 1
            monitors['global_epoch'] += 1

            # evaluate the training and validation accuracy
            train_preds, train_target, train_loss = self._predict(train_loader)
            train_acc = self.get_acc(train_preds, train_target, ncls)
            val_preds, val_target, val_loss = self._predict(val_loader)
            val_acc = self.get_acc(val_preds, val_target, ncls)
            monitors['val_inacc'] = 1 - val_acc
            monitors['val_loss'] = val_loss

            # store loss and acc
            writer.add_scalars('train', {'loss': train_loss, 'acc': train_acc}, 
                                monitors['global_epoch'])
            writer.add_scalars('val', {'loss': val_loss, 'acc': val_acc}, 
                                monitors['global_epoch'])
            # print the epoch info
            loger.info(f'-->Epoch : {monitors["epoch"]}')
            loger.info(f'  \u21b3train Loss/Acc = {train_loss:.4f}/{train_acc:.4f}'
                       f' | val Loss/Acc = {val_loss:.4f}/{val_acc:.4f}')

            # select best model on Stage 1
            if load_best_state and monitors[var_check] <= best_var:
                best_var = monitors[var_check]
                best_net_param = deepcopy(self.net.state_dict())
                best_optim_param = deepcopy(self.optimizer.state_dict())
                monitors['best_train_loss'] = train_loss
                monitors['best_epoch'] = monitors['epoch']

            # check whether to stop training
            if stopcri(monitors):
                # check whether to enter the second stage of training
                if second_stage and not early_stop_reached:
                    early_stop_reached = True
                    epoch = monitors['epoch']

                    # load the best state
                    if load_best_state:
                        self.net.load_state_dict(best_net_param)
                        self.optimizer.load_state_dict(best_optim_param)
                        train_loss = monitors['best_train_loss']
                        epoch = monitors['best_epoch']

                    loger.info('[Early Stopping Reached] -> Training on full set.')
                    loger.info(f'[Epoch = {epoch} | Loss = {train_loss:.4f}]')

                    # combine the train and val dataset
                    train_loader = self._data_loader(trainset, valset)

                    # update stop monitor and epoch
                    stopcri = ComposeStopCriteria({'Or': {
                        'cri1': {'MaxEpoch': {
                            'max_epochs': max_epochs_s2, 'var_name': 'epoch'
                        }},
                        'cri2': {'Smaller': {
                            'var': train_loss, 'var_name': 'val_loss'
                        }}
                    }})
                    self.train_details['fit']['stopcri_2'] = str(stopcri)
                    monitors['epoch'] = 0
                    load_best_state = False
                elif second_stage and early_stop_reached:
                    do_stop = True
                    best_net_param = deepcopy(self.net.state_dict())
                # no second stage
                else:
                    do_stop = True
                    if not load_best_state:
                        best_net_param = deepcopy(self.net.state_dict())

        writer.close()

        # report the checkpoint time of end and compute cost time
        loger.info(f'[Train Finish] - [{self.timer.ctime()}]')
        h, m, s = self.timer.stop()
        loger.info(f'Cost Time = {h}H:{m}M:{s:.2f}S')

        # load the best model and evaulate this model in testset
        self.net.load_state_dict(best_net_param)

        results = {}
        train_preds, train_target, train_loss = self._predict(train_loader)
        train_acc = self.get_acc(train_preds, train_target, ncls)
        results['train'] = {
            'preds': train_preds, 'target': train_target, 'acc': train_acc
        }
        val_preds, val_target, val_loss = self._predict(val_loader)
        val_acc = self.get_acc(val_preds, val_target, ncls)
        results['val'] = {
            'preds': val_preds, 'target': val_target, 'acc': val_acc
        }
        test_preds, test_target, test_loss = self._predict(test_loader)
        test_acc = self.get_acc(test_preds, test_target, ncls)
        results['test'] = {
            'preds': test_preds, 'target': test_target, 'acc': test_acc
        }

        loger.info(f'Loss: train={train_loss:.4f} | val={val_loss:.4f} | '
                   f'test={test_loss:.4f}')
        loger.info(f'Acc:  train={train_acc:.4f} | val={val_acc:.4f} | '
                   f'test={test_acc:.4f}')

        # save the training results
        self.train_details['results'] = results
        self.train_details['best_net_param'] = best_net_param

        # store the training details
        train_details_path = os.path.join(log_dir, f'train_details.pt')
        torch.save(self.train_details, train_details_path)

        # store the best model
        best_checkpoiont_path = os.path.join(log_dir, f'best_checkpoint.pth')
        torch.save(best_net_param, best_checkpoiont_path)

        return results

    def __repr__(self) -> str:
        '''Trainer details.
        '''
        s = '[Network architecture]:\n' + self.net_arch + '\n'
        s += f'[Loss function]: {self.loss_fn_type}\n'
        if self.loss_fn_args:
            s += f'[loss_fn Args]: {self.lr_sch_args}\n'
        s += f'[Optimizer]: {self.optimizer_type}\n'
        s += f'[Learning rate]: {self.lr}\n'
        if self.optimizer_args:
            s += f'[Optim Args]: {self.optimizer_args}\n'
        if self.lr_sch_type:
            s += f'[Lr scheduler]: {self.lr_sch_type}\n'
            if self.lr_sch_args:
                s += f'[lr_sch Args]: {self.lr_sch_args}\n'
        s += f'[Grad Acc]: {self.grad_acc}\n'
        s += f'[Batch Size]: {self.batch_size}\n'
        return s



from .experiments import (
    KFold,
    Holdout,
    LOSO_HO,
    LOSO_CV,
)
from .classifier import (
    TrainClassifier,
)
from .stopcriteria import (
    Criteria,
    ComposeStopCriteria,
    And,
    Or,
    MaxEpoch,
    NoDecrease,
    Bigger,
    Smaller,
)
from .evaluate import (
    save_cm_img,
)


#!/usr/bin/env python
# coding: utf-8

"""
    Basic method of `transforms.py` file.
    
    @Author  : SheepTAO
    @Time    : 2023-09-14
"""


import os
import torch
import numpy as np
import scipy.signal as signal
from numpy import ndarray
from torch import Tensor
from typing import Optional, Union, Tuple, List, Literal

from ..utils import loger, verbose, DPEEG_SEED
from ..tools.logger import _Level


def check_sub_data(sub : int, sub_data) -> bool:
    '''Check the subject's data format.

    Check whether the subject's data format meets the requirements. If it does
    not meet the requirements, an error will be thrown. If there is no format
    error, a bool value is returned. True means that the dataset has been split,
    False means that the dataset has not been split yet.

    Parameters
    ----------
    sub : int
        Number of subject.
    sub_data : Any
        Subject's data.

    Returns
    -------
    bool
        True means splited, False means un-splited.
    '''
    # splited dataset
    if isinstance(sub_data, dict):
        split = True

        # trainset and testset
        if not sorted(sub_data.keys()) == ['test', 'train']:
            raise ValueError(f'Subject {sub} should have train and test set.')

        # data and label
        for mode, dl in sub_data.items(): # type: ignore
            if not (isinstance(dl, (list, tuple)) and len(dl) == 2):
                raise ValueError(f'Subject {sub} {mode}set miss data or label.')
    # un-splited dataset
    else:
        split = False

        # data and label
        if not(isinstance(sub_data, (list, tuple)) and len(sub_data) == 2):
            raise ValueError(f'Subject {sub} miss data or label.')

    return split


def check_dataset(dataset) -> bool:
    '''Check the format of the dataset.

    Determine whether the format of the data set meets the requirements of the 
    `dpeeg` data set format. If the requirements are not met, the corresponding
    error will be thrown. If the requirements are met, a bool value will be 
    returned. True means that the dataset has been split, False means that the 
    dataset has not been split yet.

    Parameters
    ----------
    dataset : Any
        Input dataset.

    Returns
    -------
    bool
        True means splited, False means un-splited.
    '''
    # entire dataset
    if not isinstance(dataset, dict):
        raise ValueError('Dataset should be dict.')

    if not all(isinstance(k, int) for k in dataset.keys()):
        raise ValueError('Name of subject should be int.')

    split = None
    for sub, sub_data in dataset.items():
        last_sub_split = split

        split = check_sub_data(sub, sub_data)

        if last_sub_split is not None and last_sub_split != split:
            raise ValueError(f'Subject {sub} format is not uniform.')

    return split # type: ignore


def total_trials(
    dataset : dict, 
    sub : Optional[int] = None,
    mode : Literal['train', 'test', 'all'] = 'all'
) -> int:
    '''All trials in the dataset.

    Parameters
    ----------
    dataset : dict
        The dataset to be counted.
    sub : int, optional
        If None, count all subjects in the dataset.
    mode : str
        Mode to count data. If 'all', both trainset and testset will be counted.
        If 'train', only trainset will be counted. 'test' is the same.

    Returns
    -------
    int
        Total trials.
    '''
    train_trials = test_trials = 0

    for sub_id, sub_data in dataset.items():
        if sub is not None and sub_id == sub:
            train_trials += sub_data['train'][1].shape[0]
            test_trials += sub_data['test'][1].shape[0]

    if mode == 'train':
        return train_trials
    elif mode == 'test':
        return test_trials
    elif mode == 'all':
        return train_trials + test_trials
    else:
        raise KeyError(f'Only support `all`, `train` and `test` mode.')


@verbose
def split_train_test(
    *arrs, 
    test_size : float = .25, 
    seed : int = DPEEG_SEED, 
    sample : Optional[List[int]] = None, 
    verbose : _Level = None
) -> list:
    '''Split an dataset into training and testing sets. The axis along which
    to split is 0.

    Parameters
    ----------
    *arrs : sequence of indexables with same length / shape[0]
        Allowed inputs are lists and numpy arrays.
    test_size : float
        The proportion of the test set. If index is not None, test_size will be
        ignored. Default use stratified fashion and the last arr serves as the
        class labels.
    seed : int
        Random seed when splitting.
    sample : list of int, optional
        A list of integers, the entries indicate which data were selected
        as the test set. If None, test_size will be used.

    Returns
    -------
    splitting : list, length=2 * len(arrays)
        List containing train-test split of inputs.
    '''
    num_arrs = len(arrs)
    if num_arrs == 0:
        raise ValueError('At least one array required as input.')

    arr_list = [np.array(arr) for arr in arrs]
    lengths = [len(arr) for arr in arr_list]
    uniques = np.unique(lengths)
    if len(uniques) > 1:
        raise ValueError(
            'Found input variables with inconsistent numbers of samples: %r'
            % [int(l) for l in lengths]
        )
        
    if sample is not None:
        smparr = np.array(sample)
        if np.unique(smparr).size != smparr.size:
            raise IndexError(
                f'Found repeated sampling of test set: {smparr.tolist()}.'
            )
        length = lengths[0]
        if smparr.size >= uniques:
            raise IndexError(f'The number of samples (={len(smparr)}) in the '
                             'test set cannot exceed the total number of data '
                             f'sets (={length}).'
            )

        testSample, trainSample = smparr, np.setdiff1d(np.arange(length), smparr)
        res = []
        for arr in arr_list:
            res.extend([arr[trainSample], arr[testSample]])
        return res
    else:
        from sklearn.model_selection import train_test_split
        return train_test_split(
            *arr_list, test_size=test_size, random_state=seed, 
            stratify=arr_list[-1]
        )


@verbose
def merge_train_test(
    *arrs,
    verbose : _Level = None
) -> Tuple[ndarray, ndarray]:
    '''Merge the data and labels of the training set and test set.

    Parameters
    ----------
    *arrs : sequence of data (ndarray (N, ...)) and labels (ndarray (N,))
        Sequence consisting of each piece of data and its corresponding label.

    Returns
    -------
    data, label : ndarray, ndarray
        Merged data and label.

    Examples
    --------
    >>> data = np.arange(6).reshape(2, 3)
    >>> label = np.arange(2)
    >>> train, test = [data, label], [data, label]
    >>> merge_train_test(train, test)
    (array([[3, 4, 5],
            [0, 1, 2],
            [0, 1, 2],
            [3, 4, 5]]),
    array([1, 0, 0, 1]))
    '''
    if len(arrs) == 0:
        raise ValueError('At least one array required as input.')

    merge_data = []
    merge_label = []
    for i, arr in enumerate(arrs):
        if not (isinstance(arr, (list, tuple)) and len(arr) == 2):
            raise ValueError(f'Missing data or label in {i}th group.')

        data, label = arr[0], arr[1]
        if data.shape[0] != label.shape[0]:
            loger.warning(f'Length of data and label in {i}th group is not uniform.')
        merge_data.append(data)
        merge_label.append(label)

    merge_data = np.concatenate(merge_data)
    merge_label = np.concatenate(merge_label)
    shuffle = np.random.permutation(merge_label.shape[0])
    merge_data = merge_data[shuffle]
    merge_label = merge_label[shuffle]

    return merge_data, merge_label
    

@verbose
def to_tensor(
    data: Union[Tensor, ndarray],
    label: Union[Tensor, ndarray],
    verbose : _Level = None
) -> Tuple[Tensor, Tensor]:
    '''Convert the numpy data and label into trainable Tensor format.
    '''
    # https://discuss.pytorch.org/t/torch-from-numpy-not-support-negative-strides/3663
    dataT = torch.from_numpy(np.ascontiguousarray(data)).float() \
        if isinstance(data, ndarray) else data.float()
    labelT = torch.from_numpy(np.ascontiguousarray(label)).long() \
        if isinstance(label, ndarray) else label.long()
    if dataT.size(0) != labelT.size(0):
        loger.warning('Data and label do not match in the first dimension: '
                      f'{dataT.size(0)} and {labelT.size(0)} respectively.')
    return dataT, labelT


@verbose
def slide_win(
    data : ndarray,
    win : int, 
    overlap : int = 0,
    label : Optional[ndarray] = None,
    verbose : _Level = None,
) -> Tuple[ndarray, Optional[ndarray]]:
    '''This transform is only splits the time series (dim = -1) through the 
    sliding window operation on the original dataset. If the time axis is not
    divisible by the sliding window, the last remaining time data will be 
    discarded.

    Parameters
    ----------
    data : ndarray (N, ..., times)
        The data to split.
    win : int
        The size of the sliding window.
    overlap : int
        The amount of overlap between adjacent sliding windows.
    label : ndarray (N,), optional
        The label of the data. If not None, label will update with sliding window.

    Returns
    -------
    data : ndarray
        Data after sliding window.
    label : ndarray, None
        If label is None, return None. The label corresponding to each window.
    '''
    if win < 0 or overlap < 0:
        raise ValueError('win and overlap only allow positive numbers, '
                         f'but got {win} and {overlap} respectively.')

    if overlap >= win:
        raise ValueError(f'overlap={overlap} should be less than win={win}.')

    if isinstance(label, ndarray) and len(data) != len(label):
        raise ValueError('The number of label and data must be the same.')    

    loger.info(f'Sliding window with win:{win} and overlap:{overlap}.')

    end = win
    times = data.shape[-1]
    if end > times:
        loger.warning('The window is larger than the times to be split.')
        return data, label

    sld_num = 0
    data_list = []
    while end <= times:
        data_list.append(data[..., end-win:end])
        loger.info(f' Intercept data from {end-win} to {end}.')
        sld_num += 1
        end += win - overlap

    data = np.concatenate(data_list)
    shuffle = np.random.permutation(len(data))
    data = data[shuffle]
    if isinstance(label, ndarray):
        label = np.repeat(label, sld_num)
        label = label[shuffle]
    return data, label


@verbose
def segmentation_and_reconstruction(
    data : ndarray,
    label : ndarray,
    multiply : int = 1,
    n : int = 8,
) -> Tuple[ndarray, ndarray]:
    '''Signal Segmentation and Recombination in Time Domain.

    This approach is to first divide each training EEG trial into several 
    segments and then generate new artificial trials as a concatenation of 
    segments coming from different and randomly selected training trials from
    the same class while maintaining the original time order.

    Parameters
    ----------
    data : ndarray
        Data that will be segmented and randomly recombined. Shape as `(N, ..., 
        T)`, with `N` the number of data and `T` the number of samples.
    label : ndarray
        The label corresponding to the data. Shape as `(N)`.
    multiply : int
        The data will be enhanced several times.
    n : int
        The data will be segmented into n parts in time domain. And n should be
        evenly divisible by the time length of the data. eg, 250Hz data has 
        1,000 sampling points, and cutting it into 8 parts means that each seg-
        ment is 0.5s of data.

    Notes
    -----
    The function does not judge the multiples or the number of segmentation.

    References
    ----------
    F. Lotte, “Signal Processing Approaches to Minimize or Suppress 
    Calibration Time in Oscillatory Activity-Based Brain-Computer Interfaces,
    ” Proc. IEEE, vol. 103, no. 6, pp. 871-890, Jun. 2015, 
    doi: 10.1109/JPROC.2015.2404941.
    '''
    if n == 0:
        raise ValueError('Parameter n must be at least 1.')
    
    if data.shape[-1] % n:
        raise ValueError(f'The data time length ({data.shape[-1]}) should be '
                         f'a multiple of n={n}.')

    win = data.shape[-1] // n
    aug_data, aug_label = [], []

    for lb in np.unique(label):
        idx = np.where(label == lb)
        tmp_data, tmp_label = data[idx], label[idx]
        m = tmp_data.shape[0] * multiply

        tmp_aug_data = np.empty((m, *data.shape[1:]))
        for i in range(m):
            for j in range(n):
                randIdx = np.random.randint(0, tmp_data.shape[0], n)
                tmp_aug_data[i, ..., j * win : (j + 1) * win] = \
                    tmp_data[randIdx[j] , ..., j * win : (j + 1) * win]

        aug_data.append(tmp_aug_data)
        aug_label.append(np.repeat(lb, m))
        aug_data.append(tmp_data)
        aug_label.append(tmp_label)

    aug_data = np.concatenate(aug_data)
    aug_label = np.concatenate(aug_label)
    shuffle = np.random.permutation(aug_data.shape[0])
    aug_data = aug_data[shuffle]
    aug_label = aug_label[shuffle]

    return aug_data, aug_label


@verbose
def save(
    folder : str,
    input : dict,
    verbose : _Level = None,
) -> None:
    '''Save transformed dataset to a binary file in NumPy `.npy` format.

    Parameters
    ----------
    folder : str
        Folder name to save transformed data.
    input : dict
        Data to be saved.
    '''
    folder = os.path.abspath(folder)
    os.makedirs(folder, exist_ok=True)
    if os.listdir(folder):
        raise FileExistsError(f'\'{folder}\' is not a empty folder.')
    
    loger.info(f'Transformed data will be saved in: \'{folder}\'')
    for sub, data in input.items():
        loger.info(f'Save transformed data of sub_{sub}.')
        for mode in ['train', 'test']:
            if not isinstance(data[mode], (tuple, list)):
                loger.warning(f'sub{sub}_{mode} may not exist data or label.')

            file_name = os.path.join(folder, f'{sub}_{mode}')
            np.save(file_name + '_data', data[mode][0])
            np.save(file_name + '_label', data[mode][1])


@verbose
def load(
    folder : str,
    subjects : Optional[List[int]] = None,
    mode : Literal['train', 'test', 'all'] = 'all',
    verbose : _Level = None,
) -> dict:
    '''Load saved transformed dataset from folder.

    Parameters
    ----------
    folder : str
        Folder name where transformed data is saved.
    subjects : list of int, optional
        List of subject number. If None, all subjects will be loaded.
    mode : str
        Mode to load data. If 'all', both trainset and testset will be loaded. 
        If 'train', only trainset will be loaded. 'test' is the same.
    '''
    path = os.path.abspath(folder)
    loger.info(f'Loading dataset from \'{path}\'')

    path_list = os.listdir(path)
    sub_list = list(set([int(p.split('_')[0]) for p in path_list]))
    sub_list.sort()
    if subjects:
        intersection = set(subjects) & set(sub_list)
        exclude = set(subjects) - set(sub_list)
        if exclude:
            loger.warning(f'Could not find subjects: {exclude}, ' +
                          f'only load subjects: {intersection}')
        sub_list = list(intersection)

    dataset = {}
    for sub in sub_list:
        loger.info(f'Loading subject {sub}')
        file_name = os.path.join(path, str(sub))

        if mode.lower() == 'train' or mode.lower() == 'all':
            data = np.load(file_name + '_train_data.npy', allow_pickle=True)
            label = np.load(file_name + '_train_label.npy', allow_pickle=True)
            dataset.setdefault(sub, {})['train'] = [data, label]

        if mode.lower() == 'test' or mode.lower() == 'all':
            data = np.load(file_name + '_test_data.npy', allow_pickle=True)
            label = np.load(file_name + '_test_label.npy', allow_pickle=True)
            dataset.setdefault(sub, {})['test'] = [data, label]

        if mode not in ['train', 'test', 'all']:
            raise KeyError(f'Only support `all`, `train` and `test` mode.')
    loger.info('Load dataset done.')
    return dataset


@verbose
def cheby2_filter(
    data : ndarray,
    freq : float,
    l_freq : Optional[float] = None,
    h_freq : Optional[float] = None,
    transition_bandwidth : float = 2.,
    gstop : float = 30,
    gpass : float = 3,
    filter_type : Literal['filter', 'filtfilt'] = 'filter',
    verbose : _Level = None
):
    '''Filter a signal using cheby2 iir filtering.

    Parameters
    ----------
    data : ndarray
        Filtering is performed in the last dimension (time dimension).
    freq : float
        Data sampling frequency.
    l_freq, h_freq : float, optional
        Low and high cut off frequency in hertz. If l_freq is None, the data 
        are only low-passed. If h_freq is None,  the data are only high-passed.
        Both parameters cannot be None at the same time.
    transition_bandwidth : float
        The bandwidth (in hertz) of the transition region of the frequency res-
        ponse from the passband to the stopband.
    gstop : float
        The minimum attenuation in the stopband (dB).
    gpass : float
        The maximum loss in the passband (dB).
    filter_type : str
        Filter type, available options are 'filtfilt' and 'filter'.

    Returns
    -------
    data : ndarray
        Data after applying bandpass filter.
    '''
    n_freq = freq / 2   # Nyquist frequency

    if (l_freq == 0 or l_freq == None) and (h_freq == None or h_freq >= n_freq):
        loger.warning('Not doing any filtering. Invalid cut-off freq.')
        return data

    # low-passed filter
    elif l_freq == 0 or l_freq == None:
        loger.info(f'{h_freq} Hz lowpass filter.')
        fpass = h_freq / n_freq                                 # type:ignore
        fstop = (h_freq + transition_bandwidth) / n_freq        # type:ignore
        N, Ws = signal.cheb2ord(fpass, fstop, gpass, gstop)
        b, a  = signal.cheby2(N, gstop, fstop, 'lowpass')

    # high-passed filter
    elif h_freq == None or h_freq == n_freq:
        loger.info(f'{l_freq} Hz highpass filter.')
        fpass = l_freq / n_freq
        fstop = (l_freq - transition_bandwidth) / n_freq
        N, Ws = signal.cheb2ord(fpass, fstop, gpass, gstop)
        b, a  = signal.cheby2(N, gstop, fstop, 'highpass')

    # band-passed filter
    else:
        loger.info(f'{l_freq} - {h_freq} Hz bandpass filter.')
        fpass = [l_freq / n_freq, h_freq / n_freq]
        fstop = [(l_freq - transition_bandwidth) / n_freq, 
                 (h_freq + transition_bandwidth) / n_freq]
        N, Ws = signal.cheb2ord(fpass, fstop, gpass, gstop)
        b, a  = signal.cheby2(N, gstop, fstop, 'bandpass')

    if filter_type == 'filtfilt':
        out = signal.filtfilt(b, a, data)
    elif filter_type:
        out = signal.lfilter(b, a, data)
    else:
        raise ValueError(f'Unsupported filter {filter_type}.')

    return out



#!/usr/bin/env python
# coding: utf-8

"""
    This module is used to evaluate the metrics of the model and save.

    @author: SheepTAO
    @data: 2023-4-13
"""


import torch
import seaborn as sns
import matplotlib.pyplot as plt
from torch import Tensor
from torch.nn import Module
from torchinfo import summary
from torch.utils.hooks import RemovableHandle
from typing import Union, Optional, Dict, Tuple, List
from torchmetrics.functional.classification.confusion_matrix import confusion_matrix


def save_cm_img(
    preds : Tensor,
    target : Tensor,
    cls_name : Union[List, tuple],
    fig_path : str,
) -> None:
    '''Calculate and save the corresponding confusion matrix figure to the given
    img path.

    Parameters
    ----------
    preds : Tensor
        Predicted labels, as returned by a classifier.
    target : Tensor
        Ground truth (correct) labels.
    cls_name : list, tuple
        The name of dataset labels.
    fig_path : str
        Path to store the figure.
    '''
    cm = confusion_matrix(preds, target, 'multiclass', num_classes=len(cls_name))
    ax = sns.heatmap(cm / cm.sum(1, keepdim=True), annot=True, cmap='Blues',
                      xticklabels=cls_name, yticklabels=cls_name) # type: ignore
    ax.set_ylabel('Actual')
    ax.set_xlabel('Predicted')
    plt.savefig(fig_path)
    # # !!! Clear the current Axes
    # plt.clf()
    # Clear all
    plt.close()


def model_predict(
    model : Module, 
    data : Tensor,
    event_id : Optional[dict[str, int]] = None,
    no_grad : bool = True,
) -> Union[Tuple[Tensor, List[str]], Tuple[Tensor, None]]:
    '''Compute model predictions.

    Parameters
    ----------
    model : Module
        Inherit Module and should define the forward method. The first 
        parameter returned by model forward propagation is the prediction.
    data : Tensor (N, ...)
        Data to be predicted.
    event_id : dict, optional
        A dictionary containing label and corresponding id.
    no_grad : bool
        Whether to turn off autograd calculation graph recording.

    Returns
    -------
    preds : Tensor (N, ...)
        Prediction corresponding to the input data.
    labels : list of str
        If event_id not None, corresponding predicted label will be returned.
    '''
    model.eval()
    if no_grad:
        with torch.no_grad():
            out = model(data)
    else:
        out = model(data)

    out = out[0] if isinstance(out, tuple) else out
    preds = torch.argmax(out, dim=1).detach()

    if event_id:
        reversed_event_id = {v : k for k, v in event_id.items()}
        labels = [reversed_event_id[pred] for pred in preds.tolist()]
        return preds, labels
    else:
        return preds, None


def model_depth(model : Module) -> int:
    '''Compute the maximum depth of the model.
    '''
    names = dict(model.named_modules()).keys()
    name_depth = [len(name.split('.')) for name in names]
    max_depth = max(name_depth)
    return max_depth


def model_summary(model : Module):
    '''Output the model structure according to the model depth.

    This function is usually used in conjunction with the `Activation` class. 
    You can use this function to view the specific inter-layer relationships of
    the model, and then obtain the corresponding middle layer name by adding '.'
    to the inter-layer name as the names parameter of the Activation class.

    Examples
    --------
    >>> class MyModel(nn.Module):
    ...     def __init__(self) -> None:
    ...         super().__init__()
    ...         self.conv = nn.Sequential(
    ...             nn.Conv2d(3, 10, 3),
    ...             nn.Conv2d(10, 20, 4)
    ...         )
    ... 
    ...     def forward(self, x):
    ...         return self.conv(x)
    ... 
    >>> model = MyModel()
    >>> print(model_summary(model))
    MyModel (MyModel)                        --
    ├─Sequential (conv)                      --
    │    └─Conv2d (0)                        280
    │    └─Conv2d (1)                        3,220

    Then you can use `conv.0` and `conv.1` as the name of intermediate layer.
    '''
    s = summary(
        model=model,
        depth=model_depth(model),
        row_settings=['var_names'],
        verbose=0
    )
    return s


class Activation:
    def __init__(self, model : Module, names : List[str]) -> None:
        '''Get model's intermediate result.

        Allows the context manager to be used to obtain the forward propagation
        output feature map of any intermediate layer of the model.

        Parameters
        ----------
        model : Module
            Obtain the intermediate layer output of the model.
        names : list of str
            A list of intermediate module names whose outputs will be obtained.
            You can get all module names through `model.named_modules()` or 
            `dpeeg.base.evaluate.model_summary`.

        Examples
        --------
        >>> with Activation(net, ['conv.0', 'conv.1']) as act:
        >>>     model_predict(net, data)
        >>>     actmaps = act.get_actmaps()

        you can also manually remove the handler registered on the model

        >>> act = Activation(net, ['conv.0', 'conv.1'])
        >>> model_predict(net, data)
        >>> actmaps = act.get_actmaps()
        >>> act.close()
        '''        
        self.names = names
        self.handle : List[RemovableHandle] = list()
        self.activation_maps  = dict()

        for name in self.names:
            module = model.get_submodule(name)
            handle = module.register_forward_hook(self._hook_func(name))
            self.handle.append(handle)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def _hook_func(self, name):
        def hook(model, input, output):
            if isinstance(output, tuple):
                actmaps = list()
                for out in output:
                    if isinstance(out, Tensor):
                        actmaps.append(out.detach())
                    else:
                        actmaps.append(out)
                actmap = tuple(actmaps)
            elif isinstance(output, Tensor):
                actmap = output.detach()
            else:
                raise TypeError(f'Unsupport type {type(output)} as output, in '
                                f'{model._get_name()}.')
            self.activation_maps[name] = actmap
        return hook

    def get_actmaps(self) -> Dict[str, Tensor]:
        '''Get intermediate activation maps.

        Returns
        -------
        dict
            Returns the name of the intermediate layer and its corresponding
            activation feature maps.
        '''
        return self.activation_maps

    def close(self):
        '''Remove the handler and clear the activation maps.
        '''
        for handle in self.handle:
            handle.remove()
        self.handle.clear()


def get_data_by_label(data : Tensor, labels : Tensor, label : int) -> Tensor:
    '''Index data based on specified label.

    Parameters
    ----------
    data : Tensor (N, ...)
        Input data.
    labels : Tensor (N,)
        Label table corresponding to the data.
    label : int
        The label to get.

    Returns
    -------
    Tensor
        All data for the specified label.
    '''
    if label not in labels:
        raise ValueError(f'label {label} is not in the labels.')

    indices = torch.where(labels == label)
    return data[indices]
